﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections;

namespace RDFSharp.Model
{

    /// <summary>
    /// RDFCollection represents a generic collection in the RDF model.
    /// It is made up of items, which must be all resources or all literals.
    /// </summary>
    public class RDFCollection {

        #region Properties
        /// <summary>
        /// Type of the items of the collection
        /// </summary>
        public RDFModelEnums.RDFItemTypes ItemType { get; internal set; }

        /// <summary>
        /// Subject of the collection's reification
        /// </summary>
        public RDFResource ReificationSubject { get; internal set; }

        /// <summary>
        /// List of the items collected by the collection
        /// </summary>
        internal ArrayList Items { get; set; }

        #region Utilities
        /// <summary>
        /// Count of the collection's items
        /// </summary>
        public Int32 ItemsCount {
            get { return this.Items.Count; }
        }

        /// <summary>
        /// Gets the enumerator on the collection's items for iteration
        /// </summary>
        public IEnumerator ItemsEnumerator {
            get { return this.Items.GetEnumerator(); }
        }
        #endregion

        #endregion

        #region Ctors
        /// <summary>
        /// Default ctor to build an empty collection of the given flavor
        /// </summary>
        public RDFCollection(RDFModelEnums.RDFItemTypes itemType) {
           this.ItemType                   = itemType;
           this.ReificationSubject         = new RDFResource();
           this.Items                      = new ArrayList();        
        }

        /// <summary>
        /// List-based ctor to build a collection of the given flavor with given items
        /// </summary>
        public RDFCollection(RDFModelEnums.RDFItemTypes itemType, ArrayList items): this(itemType) {
            if (items != null) {
                if (this.ItemType         == RDFModelEnums.RDFItemTypes.Resource) {
                    foreach(var item in items){
                        if(item is RDFResource){
                            this.AddItem(item);
                        }
                        else{
                            throw new RDFModelException("Cannot create RDFCollection because all the elements must be resources.");
                        }
                    }    
                }
                else if (this.ItemType    == RDFModelEnums.RDFItemTypes.Literal) {
                    foreach (var item in items) {
                        if (item is RDFLiteral) {
                            this.AddItem(item);
                        }
                        else {
                            throw new RDFModelException("Cannot create RDFCollection because all the elements must be literals.");
                        }
                    }
                }                
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Adds the given item to the collection, avoiding duplicate insertions
        /// </summary>
        public RDFCollection AddItem(Object item) {
            if (item != null) {
                //Try to add a resource
                if (item is RDFResource && this.ItemType == RDFModelEnums.RDFItemTypes.Resource) {
                    //Sequential flag-terminating search of the item into the collection
                    IEnumerator itemEnum = this.ItemsEnumerator;
                    Boolean itemFound    = false;
                    while (!itemFound   && itemEnum.MoveNext()) {
                        if(((RDFResource)itemEnum.Current).Equals((RDFResource)item)){
                            itemFound    = true;
                        }    
                    }
                    if (!itemFound) {
                        this.Items.Add(item);
                    }
                }
                //Try to add a literal
                else if (item is RDFLiteral && this.ItemType == RDFModelEnums.RDFItemTypes.Literal) {
                    IEnumerator itemEnum = this.ItemsEnumerator;
                    Boolean itemFound    = false;
                    while (!itemFound   && itemEnum.MoveNext()) {
                        if (((RDFLiteral)itemEnum.Current).Equals((RDFLiteral)item)) {
                            itemFound    = true;
                        }
                    }
                    if (!itemFound) {
                        this.Items.Add(item);
                    }
                }
            }
            return this;
        }

        /// <summary>
        /// Removes the given item from the collection
        /// </summary>
        public RDFCollection RemoveItem(Object item) {
            if (item != null) {
                //Try to remove a resource
                if (item is RDFResource && this.ItemType == RDFModelEnums.RDFItemTypes.Resource) {
                    ArrayList resultList = new ArrayList();
                    IEnumerator itemEnum = this.ItemsEnumerator;
                    while (itemEnum.MoveNext()) {
                        if (!((RDFResource)itemEnum.Current).Equals((RDFResource)item)) {
                            resultList.Add(itemEnum.Current);
                        }
                    }
                    this.Items           = resultList;
                }
                //Try to remove a literal
                else if (item is RDFLiteral && this.ItemType == RDFModelEnums.RDFItemTypes.Literal) {
                    ArrayList resultList = new ArrayList();
                    IEnumerator itemEnum = this.ItemsEnumerator;
                    while (itemEnum.MoveNext()) {
                        if (!((RDFLiteral)itemEnum.Current).Equals((RDFLiteral)item)) {
                            resultList.Add(itemEnum.Current);
                        }
                    }
                    this.Items           = resultList;
                }
            }
            return this;
        }

        /// <summary>
        /// Removes all the items from the collection
        /// </summary>
        public RDFCollection ClearItems() {
            this.Items.Clear();
            return this;
        }

        /// <summary>
        /// Builds the reification graph of the collection
        /// </summary>
        public RDFGraph ReifyCollection() {
            RDFGraph reifiedCollection      = new RDFGraph();
            RDFResource reifSubject         = this.ReificationSubject;
            Int32 itemCount                 = 0;

            IEnumerator listEnum            = this.ItemsEnumerator;
            while (listEnum.MoveNext()) {
                //Count the items to keep track of the last one, which will be connected to rdf:nil
                itemCount++;

                //  Subject -> rdf:type  -> rdf:List
                RDFTriple t1                = new RDFTriple(reifSubject, RDFVocabulary.RDF.TYPE, RDFVocabulary.RDF.LIST);
                reifiedCollection.AddTriple(t1);

                //  Subject -> rdf:first -> RDFCollection.ITEM(I)
                RDFTriple t2                = null;
                if (this.ItemType          == RDFModelEnums.RDFItemTypes.Resource) {
                    t2                      = new RDFTriple(reifSubject, RDFVocabulary.RDF.FIRST, (RDFResource)listEnum.Current);
                }
                else {
                    t2                      = new RDFTriple(reifSubject, RDFVocabulary.RDF.FIRST, (RDFLiteral)listEnum.Current);
                }
                reifiedCollection.AddTriple(t2);

                //Not the last one: Subject -> rdf:rest  -> NEWBLANK
                RDFTriple t3                = null;
                if (itemCount < this.ItemsCount) {
                    RDFResource newSub      = new RDFResource();
                    t3                      = new RDFTriple(reifSubject, RDFVocabulary.RDF.REST, newSub);
                    reifSubject             = newSub;
                }
                //The last one: Subject -> rdf:rest  -> rdf:nil
                else {
                    t3                      = new RDFTriple(reifSubject, RDFVocabulary.RDF.REST, RDFVocabulary.RDF.NIL);
                }
                reifiedCollection.AddTriple(t3);
            }

            return reifiedCollection;
        }
        #endregion

    }

    /// <summary>
    /// RDFCollectionItem represents an item of a collection
    /// </summary>
    internal class RDFCollectionItem {

        #region Properties
        /// <summary>
        /// Type of the collection item
        /// </summary>
        internal RDFModelEnums.RDFItemTypes ItemType { get; set; }

        /// <summary>
        /// Value of the collection item
        /// </summary>
        internal Object ItemValue { get; set; }

        /// <summary>
        /// Pointer to the next item of the collection
        /// </summary>
        internal Object ItemNext { get; set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build a RDFCollectionItem with the given parameters
        /// </summary>
        internal RDFCollectionItem(RDFModelEnums.RDFItemTypes itemType, Object itemValue, Object itemNext) {
            this.ItemType  = itemType;
            this.ItemValue = itemValue;
            this.ItemNext  = itemNext;
        }
        #endregion

    }

}