﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;

namespace RDFSharp.Model
{

    /// <summary>
    /// RDFGraph represents a graph in the RDF model.
    /// </summary>
    public class RDFGraph: IEquatable<RDFGraph> {

        #region Data
        private Uri context;
        #endregion

        #region Properties
        /// <summary>
        /// Uri of the graph
        /// </summary>
        public Uri Context {
            get {
                return this.context;
            } 
            set {
                this.context = (value == null ? RDFNamespaceRegister.DefaultNamespace.Namespace : value);
            } 
        }

        /// <summary>
        /// Metadata of the graph
        /// </summary>
        internal RDFGraphMetadata GraphMetadata { get; set; }

        /// <summary>
        /// Count of the graph's triples
        /// </summary>
        public Int64 TriplesCount {
            get { return this.Triples.Count; }
        }

        /// <summary>
        /// Gets the enumerator on the graph's triples for iteration
        /// </summary>
        public IEnumerator<RDFTriple> TriplesEnumerator  {
            get { return this.Triples.Values.GetEnumerator(); }
        }

        /// <summary>
        /// List of triples embedded into the graph
        /// </summary>
        internal Dictionary<Int64, RDFTriple>  Triples { get; set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default ctor to build an empty graph
        /// </summary>
        public RDFGraph() {
            this.Context       = RDFNamespaceRegister.DefaultNamespace.Namespace;
            this.GraphMetadata = new RDFGraphMetadata();
            this.Triples       = new Dictionary<Int64, RDFTriple>();
        }

        /// <summary>
        /// List-based ctor to build a graph with the given list of triples
        /// </summary>
        public RDFGraph(List<RDFTriple> triples): this() {
            if (triples != null) {
                triples.ForEach(t => {
                    this.AddTriple(t);
                });
            }
        }
        #endregion
        
        #region Interfaces
        /// <summary>
        /// Gives the string representation of the graph
        /// </summary>
        public override String ToString() {
            return this.Context.ToString();
        }

        /// <summary>
        /// Performs the equality comparison between two graphs
        /// </summary>
        public Boolean Equals(RDFGraph other) {
            if (ReferenceEquals(this, other)) {
                return true;
            }
            if (this      == null) {
                if (other == null) {
                    return true;
                }
                else {
                    return false;
                }
            }
            else {
                if (other == null) {
                    return false;
                }
                else {
                    //Effective Comparison
                    if (this.TriplesCount != other.TriplesCount) {
                        return false;
                    }
                    var tEnum = this.TriplesEnumerator;
                    while (tEnum.MoveNext()) {
                        if (!other.Triples.ContainsKey(tEnum.Current.TripleID)) {
                            return false;
                        }
                    }
                    return true;
                }
            }
        }
        #endregion

        #region Methods

        #region Add
        /// <summary>
        /// Adds the given triple to the graph, avoiding duplicate insertions;
        /// It also updates the internal metadata of the graph
        /// </summary>
        public RDFGraph AddTriple(RDFTriple triple) {
            if (triple != null) {
                if (!this.Triples.ContainsKey(triple.TripleID)) {
                    this.Triples.Add(triple.TripleID, triple);
                    this.GraphMetadata.UpdateMetadata(triple);
                }
            }
            return this;
        }

        /// <summary>
        /// Adds the given container to the graph
        /// </summary>
        public RDFGraph AddContainer(RDFContainer container) {
            if (container             != null) {
                //Reify the container to get its graph representation
                RDFGraph reifContainer = container.ReifyContainer();
                //Iterate on the constructed triples
                var cEnum              = reifContainer.TriplesEnumerator;
                while (cEnum.MoveNext()) {
                    this.AddTriple(cEnum.Current);
                }                
            }
            return this;
        }

        /// <summary>
        /// Adds the given collection to the graph
        /// </summary>
        public RDFGraph AddCollection(RDFCollection collection) {
            if (collection             != null) {
                //Reify the collection to get its graph representation
                RDFGraph reifCollection = collection.ReifyCollection();
                //Iterate on the constructed triples
                var cEnum               = reifCollection.TriplesEnumerator;
                while (cEnum.MoveNext()) {
                    this.AddTriple(cEnum.Current);
                }
            }
            return this;
        }
        #endregion

        #region Remove
        /// <summary>
        /// Removes the given triple from the graph;
        /// It also updates the internal metadata of the graph
        /// </summary>
        public RDFGraph RemoveTriple(RDFTriple triple) {
            if (triple != null) {
                if (this.Triples.ContainsKey(triple.TripleID)) {
                    this.Triples.Remove(triple.TripleID);
                    //Backup the graph after the triple's removal operation
                    Dictionary<Int64, RDFTriple> backupGraph = new Dictionary<Int64, RDFTriple>(this.Triples);
                    //Clear the graph's triples and metadata
                    this.ClearTriples();
                    //Reconstruct the graph's triples and metadata from the backup
                    var rEnum     = backupGraph.GetEnumerator();
                    while (rEnum.MoveNext()) {
                        this.AddTriple(rEnum.Current.Value);
                    }
                }
            }
            return this;
        }

        /// <summary>
        /// Clears the triples of the graph and, consequently, all its metadata
        /// </summary>
        public RDFGraph ClearTriples() {
            this.Triples.Clear();
            this.GraphMetadata.ClearMetadata();
            return this;
        }
        #endregion

        #region Select
        /// <summary>
        /// Gets the subgraph containing triples with the specified resource as subject 
        /// </summary>
        public RDFGraph SelectTriplesBySubject(RDFResource subjectResource) {
            List<RDFTriple> matchingTriples = new List<RDFTriple>();
            if (subjectResource != null) {
                matchingTriples.AddRange(((from t in this.Triples.Values.ToList<RDFTriple>()
                                           where t.Subject.Equals(subjectResource)
                                           select t).ToList<RDFTriple>()));
            }
            return new RDFGraph(matchingTriples);
        }

        /// <summary>
        /// Gets the subgraph containing triples with the specified resource as predicate
        /// </summary>
        public RDFGraph SelectTriplesByPredicate(RDFResource predicateResource) {
            List<RDFTriple> matchingTriples = new List<RDFTriple>();
            if (predicateResource != null) {
                matchingTriples.AddRange(((from t in this.Triples.Values.ToList<RDFTriple>()
                                           where t.Predicate.Equals(predicateResource)
                                           select t).ToList<RDFTriple>()));
            }
            return new RDFGraph(matchingTriples);
        }

        /// <summary>
        /// Gets the subgraph containing triples with the specified resource as of object 
        /// </summary>
        public RDFGraph SelectTriplesByObject(RDFResource objectResource) {
            List<RDFTriple> matchingTriples = new List<RDFTriple>();
            if (objectResource != null) {
                matchingTriples.AddRange(((from t in this.Triples.Values.ToList<RDFTriple>()
                                           where t.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO && t.Object.Equals(objectResource)
                                           select t).ToList<RDFTriple>()));
            }
            return new RDFGraph(matchingTriples);
        }

        /// <summary>
        /// Gets the subgraph containing triples with the specified literal as of object 
        /// </summary>
        public RDFGraph SelectTriplesByLiteral(RDFLiteral objectLiteral) {
            List<RDFTriple> matchingTriples = new List<RDFTriple>();
            if (objectLiteral != null) {
                matchingTriples.AddRange(((from t in this.Triples.Values.ToList<RDFTriple>()
                                           where t.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPL && t.Object.Equals(objectLiteral)
                                           select t).ToList<RDFTriple>()));
            }
            return new RDFGraph(matchingTriples);
        }
        #endregion

        #region Set
        /// <summary>
        /// Builds a new intersection graph from this graph and a given one
        /// </summary>
        public RDFGraph IntersectWith(RDFGraph graph) {
            RDFGraph result = new RDFGraph();
            if (graph      != null) {
                var gEnum   = graph.TriplesEnumerator;
                while (gEnum.MoveNext()) {
                    if (this.Triples.ContainsKey(gEnum.Current.TripleID)) {
                        result.AddTriple(gEnum.Current);
                    }
                }
            }
            return result;
        }

        /// <summary>
        /// Builds a new union graph from this graph and a given one
        /// </summary>
        public RDFGraph UnionWith(RDFGraph graph) {
            RDFGraph result = new RDFGraph(this.Triples.Values.ToList<RDFTriple>());
            if (graph      != null) {
                var gEnum   = graph.TriplesEnumerator;
                while (gEnum.MoveNext()) {
                    result.AddTriple(gEnum.Current);
                }
            }
            return result;
        }

        /// <summary>
        /// Builds a new difference graph from this graph and a given one
        /// </summary>
        public RDFGraph DifferenceWith(RDFGraph graph) {
            RDFGraph result = null;
            if (graph      != null) {
                result      = new RDFGraph();
                var tEnum   = this.TriplesEnumerator;
                while (tEnum.MoveNext()) {
                    if (!graph.Triples.ContainsKey(tEnum.Current.TripleID)) {
                        result.AddTriple(tEnum.Current);
                    }
                }
            }
            else {
                result      = this;
            }
            return result;
        }
        #endregion

        #endregion

    }

}