﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;

namespace RDFSharp.Model
{

    /// <summary>
    /// RDFNamespaceRegister is a singleton container for registered RDF namespaces.
    /// </summary>
    public sealed class RDFNamespaceRegister {

        #region Properties
        /// <summary>
        /// Singleton centralized instance of the RDFNamespaceRegister class
        /// </summary>
        internal static RDFNamespaceRegister Instance { get; set; }

        /// <summary>
        /// List of registered namespaces
        /// </summary>
        internal List<RDFNamespace> Register { get; set; }

        #region 
        /// <summary>
        /// Default namespace of the library
        /// </summary>
        public static RDFNamespace DefaultNamespace { get; internal set; }

        /// <summary>
        /// Count of the register's namespaces
        /// </summary>
        public static Int32 NamespacesCount {
            get { return RDFNamespaceRegister.Instance.Register.Count; }
        }

        /// <summary>
        /// Gets the enumerator on the register's namespaces for iteration
        /// </summary>
        public static IEnumerator<RDFNamespace> NamespacesEnumerator {
            get { return RDFNamespaceRegister.Instance.Register.GetEnumerator(); }
        }
        #endregion

        #endregion

        #region Ctors
        /// <summary>
        /// Default ctor to initialize the singleton instance of the class
        /// and to preload the most common namespaces into the register
        /// </summary>
        static RDFNamespaceRegister() {
            RDFNamespaceRegister.Instance                 = new RDFNamespaceRegister();
            RDFNamespaceRegister.Instance.Register        = new List<RDFNamespace>();
            
            //Default namespace
            RDFNamespaceRegister.DefaultNamespace         = new RDFNamespace("rdfsharp", "http://rdfsharp.codeplex.com/default_graph#");
            RDFNamespaceRegister.AddNamespace(RDFNamespaceRegister.DefaultNamespace);

            //Basic namespaces
            //xsd
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.XSD.PREFIX,     RDFVocabulary.XSD.BASE_URI));
            //xml
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.XML.PREFIX,     RDFVocabulary.XML.BASE_URI));
            //rdf
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.RDF.PREFIX,     RDFVocabulary.RDF.BASE_URI));
            //rdfs
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.RDFS.PREFIX,    RDFVocabulary.RDFS.BASE_URI));
            //owl
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.OWL.PREFIX,     RDFVocabulary.OWL.BASE_URI));
            
            //Extended namespaces
            //foaf
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.FOAF.PREFIX,    RDFVocabulary.FOAF.BASE_URI));
            //skos
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.SKOS.PREFIX,    RDFVocabulary.SKOS.BASE_URI));
            //dc
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.DC.PREFIX,      RDFVocabulary.DC.BASE_URI));
            //geo
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.GEO.PREFIX,     RDFVocabulary.GEO.BASE_URI));
			//dbpedia
            RDFNamespaceRegister.AddNamespace(new RDFNamespace(RDFVocabulary.DBPEDIA.PREFIX, RDFVocabulary.DBPEDIA.BASE_URI));
        }
        #endregion

        #region Methods
        /// <summary>
        /// Sets the given namespace as new default namespace for the library 
        /// </summary>
        public static void SetDefaultNamespace(RDFNamespace defaultNamespace) {
            if (defaultNamespace != null && !defaultNamespace.Equals(RDFNamespaceRegister.DefaultNamespace)) {
                RDFNamespaceRegister.DefaultNamespace = defaultNamespace;
                RDFNamespaceRegister.AddNamespace(defaultNamespace);
            }
        }

        /// <summary>
        /// Adds the given namespace to the register, avoiding duplicate insertions
        /// </summary>
        public static void AddNamespace(RDFNamespace nSpace) {
            if (nSpace != null) {
                if (!RDFNamespaceRegister.ContainsNamespace(nSpace)) {
                    RDFNamespaceRegister.Instance.Register.Add(nSpace);
                }
            }
        }

        /// <summary>
        /// Removes the given namespace from the register
        /// </summary>
        public static void RemoveNamespace(RDFNamespace nSpace) {
            if (nSpace != null && !nSpace.Equals(RDFNamespaceRegister.DefaultNamespace)) { //DefaultNamespace can't be removed
                RDFNamespaceRegister.Instance.Register.RemoveAll(ns => (ns.Prefix.Equals(nSpace.Prefix)) || ns.Namespace.Equals(nSpace.Namespace));
            }
        }

        /// <summary>
        /// Checks for existence of the given namespace in the register by seeking presence of its prefix or its uri
        /// </summary>
        public static Boolean ContainsNamespace(RDFNamespace nSpace) {
            if (nSpace != null) {
                return RDFNamespaceRegister.Instance.Register.Exists(ns => (ns.Prefix.Equals(nSpace.Prefix)) || ns.Namespace.Equals(nSpace.Namespace));
            }
            else {
                return false;
            }
        }

        /// <summary>
        /// Retrieves a namespace from the register by seeking presence of its uri
        /// </summary>
        public static RDFNamespace GetByNamespace(String nSpace) {
            Uri tempNS = RDFModelUtilities.GetUriFromString(nSpace);
            if(tempNS != null){
                return RDFNamespaceRegister.Instance.Register.Find(ns => ns.Namespace.Equals(tempNS));
            }
            else {
                return null;
            }
        }

        /// <summary>
        /// Retrieves a namespace from the register by seeking presence of its prefix
        /// </summary>
        public static RDFNamespace GetByPrefix(String prefix) {
            if (prefix != null && prefix.Trim() != String.Empty) {
                return RDFNamespaceRegister.Instance.Register.Find(ns => ns.Prefix.Equals(prefix));
            }
            else {
                 return null;
            }
        }
        #endregion

    }

}