﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;
using System.Web;

namespace RDFSharp.Model
{

    /// <summary>
    /// TriX is responsible for managing serialization to and from TriX data format.
    /// </summary>
    internal static class TriX {

        #region Methods
        /// <summary>
        /// Serializes the given graph to the given filepath using TriX data format. 
        /// </summary>
        internal static void Serialize(RDFGraph graph, String filepath) {
            try {

                #region serialize
                using (XmlTextWriter trixWriter = new XmlTextWriter(filepath, Encoding.UTF8))  {
                    XmlDocument trixDoc         = new XmlDocument();
                    trixWriter.Formatting       = Formatting.Indented;

                    #region xmlDecl
                    XmlDeclaration trixDecl     = trixDoc.CreateXmlDeclaration("1.0", null, null);
                    trixDoc.AppendChild(trixDecl);
                    #endregion

                    #region trixRoot
                    XmlNode trixRoot            = trixDoc.CreateNode(XmlNodeType.Element, "TriX", null);
                    XmlAttribute trixRootNS     = trixDoc.CreateAttribute("xmlns");
                    XmlText trixRootNSText      = trixDoc.CreateTextNode("http://www.w3.org/2004/03/trix/trix-1/");
                    trixRootNS.AppendChild(trixRootNSText);
                    trixRoot.Attributes.Append(trixRootNS);

                    #region graph
                    XmlNode graphElement        = trixDoc.CreateNode(XmlNodeType.Element, "graph", null);
                    XmlNode graphUriElement     = trixDoc.CreateNode(XmlNodeType.Element, "uri", null);
                    XmlText graphUriElementT    = trixDoc.CreateTextNode(graph.ToString());
                    graphUriElement.AppendChild(graphUriElementT);
                    graphElement.AppendChild(graphUriElement);

                    #region triple
                    graph.Triples.Values.ToList<RDFTriple>().ForEach(t =>  {
                        XmlNode tripleElement   = trixDoc.CreateNode(XmlNodeType.Element, "triple", null);

                        #region subj
                        XmlNode subjElement     = null;
                        XmlText subjElementText = null;
                        if (((RDFResource)t.Subject).IsBlank) {
                            subjElement         = trixDoc.CreateNode(XmlNodeType.Element, "id", null);
                            subjElementText     = trixDoc.CreateTextNode(t.Subject.ToString().Replace("bnode:", "_:"));
                        }
                        else {
                            subjElement         = trixDoc.CreateNode(XmlNodeType.Element, "uri", null);
                            subjElementText     = trixDoc.CreateTextNode(t.Subject.ToString());
                        }
                        subjElement.AppendChild(subjElementText);
                        tripleElement.AppendChild(subjElement);
                        #endregion

                        #region pred
                        XmlNode uriElementP     = trixDoc.CreateNode(XmlNodeType.Element, "uri", null);
                        XmlText uriTextP        = trixDoc.CreateTextNode(t.Predicate.ToString());
                        uriElementP.AppendChild(uriTextP);
                        tripleElement.AppendChild(uriElementP);
                        #endregion

                        #region object
                        if (t.TripleFlavor        == RDFModelEnums.RDFTripleFlavors.SPO) {
                            XmlNode objElement     = null;
                            XmlText objElementText = null;
                            if (((RDFResource)t.Object).IsBlank)  {
                                objElement         = trixDoc.CreateNode(XmlNodeType.Element, "id", null);
                                objElementText     = trixDoc.CreateTextNode(t.Object.ToString().Replace("bnode:", "_:"));
                            }
                            else  {
                                objElement         = trixDoc.CreateNode(XmlNodeType.Element, "uri", null);
                                objElementText     = trixDoc.CreateTextNode(t.Object.ToString());
                            }
                            objElement.AppendChild(objElementText);
                            tripleElement.AppendChild(objElement);
                        }
                        #endregion

                        #region literal
                        else {

                            #region plain literal
                            if (t.Object is RDFPlainLiteral) {
                                XmlNode plainLiteralElement = trixDoc.CreateNode(XmlNodeType.Element, "plainLiteral", null);
                                if (((RDFPlainLiteral)t.Object).Language != String.Empty)  {
                                    XmlAttribute xmlLang    = trixDoc.CreateAttribute(RDFVocabulary.XML.PREFIX + ":lang", RDFVocabulary.XML.BASE_URI);
                                    XmlText xmlLangText     = trixDoc.CreateTextNode(((RDFPlainLiteral)t.Object).Language);
                                    xmlLang.AppendChild(xmlLangText);
                                    plainLiteralElement.Attributes.Append(xmlLang);
                                }
                                XmlText plainLiteralText    = trixDoc.CreateTextNode(HttpUtility.HtmlDecode(((RDFLiteral)t.Object).Value));
                                plainLiteralElement.AppendChild(plainLiteralText);
                                tripleElement.AppendChild(plainLiteralElement);
                            }
                            #endregion

                            #region typed literal
                            else {
                                XmlNode typedLiteralElement = trixDoc.CreateNode(XmlNodeType.Element, "typedLiteral", null);
                                XmlAttribute datatype       = trixDoc.CreateAttribute("datatype");
                                XmlText datatypeText        = trixDoc.CreateTextNode(((RDFTypedLiteral)t.Object).Datatype.ToString());
                                datatype.AppendChild(datatypeText);
                                typedLiteralElement.Attributes.Append(datatype);
                                XmlText typedLiteralText    = trixDoc.CreateTextNode(HttpUtility.HtmlDecode(((RDFLiteral)t.Object).Value));
                                typedLiteralElement.AppendChild(typedLiteralText);
                                tripleElement.AppendChild(typedLiteralElement);
                            }
                            #endregion

                        }
                        #endregion

                        graphElement.AppendChild(tripleElement);
                    });
                    #endregion

                    trixRoot.AppendChild(graphElement);
                    #endregion

                    trixDoc.AppendChild(trixRoot);
                    #endregion

                    trixDoc.Save(trixWriter);
                }
                #endregion

            }
            catch (Exception ex) {
                throw new RDFModelException("Cannto serialize TriX because: " + ex.Message);
            }
        }

        /// <summary>
        /// Deserializes the given TriX filepath to a graph. 
        /// </summary>
        internal static RDFGraph Deserialize(String filepath) {
            try {

                #region deserialize
                XmlReaderSettings xrs       = new XmlReaderSettings();
                xrs.IgnoreComments          = true;

#if NET40
                xrs.DtdProcessing           = DtdProcessing.Ignore;
#endif
#if NET35
                xrs.ProhibitDtd             = false;
#endif


                using (StreamReader sr      = new StreamReader(filepath)) {
                    using(XmlReader xr      = XmlTextReader.Create(sr, xrs)) {
                        RDFGraph result     = new RDFGraph();

                        #region load
                        XmlDocument trixDoc = new XmlDocument();
                        trixDoc.Load(xr);
                        #endregion

                        #region graph
                        var graphEnum                   = trixDoc.DocumentElement.ChildNodes.GetEnumerator();
                        while (graphEnum.MoveNext()) {
                            XmlNode  graph              = (XmlNode)graphEnum.Current;
                            RDFGraph g                  = new RDFGraph();

                            #region triple
                            var tripleEnum              = graph.ChildNodes.GetEnumerator();
                            while (tripleEnum.MoveNext()) {
                                XmlNode triple          = (XmlNode)tripleEnum.Current;

                                //Graph is named with an Uri: there is an <uri> element preceding all <triple> elements
                                if (triple.Name.Equals("uri")) {
                                    g.Context           = RDFModelUtilities.GetUriFromString(triple.ChildNodes[0].InnerText);
                                }

                                //Graph is not named with an Uri: there is not <uri> element preceding all <triple> elements
                                else {

                                    #region subj
                                    RDFResource subj    = new RDFResource(triple.ChildNodes[0].InnerText.Replace("_:", "bnode:"));
                                    #endregion

                                    #region pred
                                    RDFResource pred    = new RDFResource(triple.ChildNodes[1].InnerText);
                                    #endregion

                                    #region object
                                    if (triple.ChildNodes[2].Name.Equals("uri") || triple.ChildNodes[2].Name.Equals("id")) {
                                        RDFResource obj = new RDFResource(triple.ChildNodes[2].InnerText.Replace("_:", "bnode:"));
                                        RDFTriple t     = new RDFTriple(subj, pred, obj);
                                        g.AddTriple(t);
                                        continue;
                                    }
                                    #endregion

                                    #region literal

                                    #region plain literal
                                    else if (triple.ChildNodes[2].Name.Equals("plainLiteral")) {
									    RDFPlainLiteral plit     = null;
									    RDFTriple t              = null;
                                        //Seek the eventual "xml:lang" attribute
									    if (triple.ChildNodes[2].Attributes.Count > 0) {
                                            XmlAttribute xmlLang = triple.ChildNodes[2].Attributes[RDFVocabulary.XML.PREFIX + ":lang"];
										    if(xmlLang != null){
											    plit             = new RDFPlainLiteral(HttpUtility.HtmlDecode(triple.ChildNodes[2].InnerText), xmlLang.Value);                                        
										    }
										    else{
											    plit             = new RDFPlainLiteral(HttpUtility.HtmlDecode(triple.ChildNodes[2].InnerText));
										    }
                                        }
                                        else {
                                            plit                 = new RDFPlainLiteral(HttpUtility.HtmlDecode(triple.ChildNodes[2].InnerText));
                                        }
									    t                        = new RDFTriple(subj, pred, plit);
                                        g.AddTriple(t);
                                        continue;
                                    }
                                    #endregion

                                    #region typed literal
                                    else if (triple.ChildNodes[2].Name.Equals("typedLiteral")) {
                                        XmlAttribute datatype     = triple.ChildNodes[2].Attributes[0];
                                        RDFTypedLiteral tlit      = new RDFTypedLiteral(HttpUtility.HtmlDecode(triple.ChildNodes[2].InnerText), RDFModelUtilities.GetDatatypeFromString(datatype.Value));
                                        RDFTriple t               = new RDFTriple(subj, pred, tlit);
                                        g.AddTriple(t);
                                        continue;
                                    }
                                    #endregion

                                    #endregion

                                }

                            }
                            #endregion

                            result                      = result.UnionWith(g);
                            result.Context              = g.Context;
                        }
                        #endregion

                        return result;
                    }
                }
                #endregion

            }
            catch (Exception ex) {
                throw new RDFModelException("Cannot deserialize TriX because: " + ex.Message);
            }
        }
        #endregion

    }

}