﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace RDFSharp.Model
{

    /// <summary>
    /// Turtle is responsible for managing serialization to Turtle data format.
    /// </summary>
    internal static class Turtle {

        #region Methods
        /// <summary>
        /// Serializes the given graph to the given filepath using Turtle data format. 
        /// </summary>
        internal static void Serialize(RDFGraph graph, String filepath) {
            try {

                #region serialize
                using (StreamWriter sw = new StreamWriter(filepath, false)) {

                    #region prefixes
                    //Sort the prefixes for readability
                    List<RDFNamespace> sortedPrefixes  = (from    pfix in graph.GraphMetadata.Namespaces
                                                          orderby pfix.Prefix
                                                          select  pfix).ToList<RDFNamespace>();

                    //Write the graph's collected prefixes and the graph's base namespace
                    sortedPrefixes.ForEach(p => {
                        sw.WriteLine("@prefix " + p.Prefix + ": <" + p.Namespace + ">.");
                    });
                    sw.WriteLine("@base <" + graph.Context + ">.\n");
                    #endregion

                    #region linq
                    //Group the graph's triples by subj and pred
                    var groupedList             = (from    triple in graph.Triples.Values.ToList<RDFTriple>()
                                                   orderby triple.Subject.ToString(), triple.Predicate.ToString()
                                                   group   triple by new {
                                                        subj = triple.Subject.ToString(),
                                                        pred = triple.Predicate.ToString()
                                                  });
                    var groupedListLast         = groupedList.Last();
                    #endregion
												   
					#region triples
                    String actualSubj           = String.Empty;
                    String abbreviatedSubj      = String.Empty;
                    String actualPred           = String.Empty;
                    String abbreviatedPred      = String.Empty;
                    String spaceConst           = " ";
                    StringBuilder result        = new StringBuilder();

					//Iterate over the calculated groups
                    foreach (var group in groupedList) {
                        var groupLast           = group.Last();

                        //Reset the flag of subj printing for the new iteration
                        Boolean subjPrint       = false;

                        #region subj
                        //New subj found: write the finished Turtle token to the file, then start collecting the new one
                        if (!actualSubj.Equals(group.Key.subj)) {
                            if (result.Length > 0) {
                                result.Replace(";", ".", result.Length - 4, 1);
                                sw.Write(result.ToString());
                                result.Remove(0, result.Length - 1);
                            }
                            actualSubj          = group.Key.subj;
                            actualPred          = String.Empty;
                            if (!actualSubj.StartsWith("_:")) {
                                //Turtle goody to abbreviate Uris which are relative to the graph Uri
                                if (!actualSubj.Contains(graph.Context.ToString())) {
                                    abbreviatedSubj = RDFModelUtilities.AbbreviateNamespace(actualSubj);
                                }
                                else {
                                    abbreviatedSubj = actualSubj.Replace(graph.Context.ToString(), ":");
                                }
                            }
                            //Optimization: do not try to resolve Uri of blank nodes, just use them
                            else {
                                abbreviatedSubj     = actualSubj;
                            }
                            result.Append(abbreviatedSubj + " ");
                            subjPrint           = true;
                        }
                        #endregion

                        #region predObjList
                        //Iterate over the triples of the current group
                        foreach (var triple in group) {

                            #region pred
                            //New pred found: collect it to the actual Turtle token.
                            if (!actualPred.Equals(triple.Predicate.ToString())) {
                                if (!subjPrint) {
                                    result.Append(spaceConst.PadRight(abbreviatedSubj.Length + 1)); //pretty-printing spaces to align the predList
                                }
                                actualPred          = triple.Predicate.ToString();
                                //Turtle goody to abbreviate Uris which are relative to the graph Uri
                                if (!actualPred.Contains(graph.Context.ToString())) {
                                    abbreviatedPred = RDFModelUtilities.AbbreviateNamespace(actualPred);
                                }
                                else {
                                    abbreviatedPred = actualPred.Replace(graph.Context.ToString(), ":");
                                }
                                //Turtle goody for "rdf:type" shortcutting to "a"
                                if (abbreviatedPred == RDFVocabulary.RDF.PREFIX + ":type") {
                                    abbreviatedPred = "a";
                                }
                                result.Append(abbreviatedPred + " ");
                            }
                            #endregion

                            #region object
                            //Collect the object or the literal to the Turtle token
                            if (triple.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO) {
                                String obj           = triple.Object.ToString();
                                if (!obj.StartsWith("_:")) {
                                    //Turtle goody to abbreviate Uris which are relative to the graph Uri
                                    if (!obj.Contains(graph.Context.ToString())) {
                                        result.Append(RDFModelUtilities.AbbreviateNamespace(obj));
                                    }
                                    else {
                                        result.Append(obj.Replace(graph.Context.ToString(), ":"));
                                    }
                                }
                                //Optimization: do not try to resolve Uri of blank nodes, just use them
                                else {
                                    result.Append(obj);
                                }
                            }
                            #endregion

                            #region literal
                            else {
                                if (triple.Object is RDFTypedLiteral) {
                                    RDFTypedLiteral tLit = new RDFTypedLiteral("\"" + ((RDFLiteral)triple.Object).Value.Replace("\"","\\\"") + "\"", ((RDFTypedLiteral)triple.Object).Datatype);
                                    result.Append(RDFModelUtilities.AbbreviateNamespace(tLit.ToString()));
                                }
                                else {
                                    RDFPlainLiteral pLit = new RDFPlainLiteral("\"" + ((RDFLiteral)triple.Object).Value.Replace("\"", "\\\"") + "\"", ((RDFPlainLiteral)triple.Object).Language);
                                    result.Append(pLit.ToString());
                                }
                            }
                            #endregion

                            #region continuation goody
                            //Then append the appropriated Turtle continuation goody ("," or ";")
                            if (!triple.Equals(groupLast)) {
                                result.Append(", ");
                            }
                            else {
                                result.AppendLine("; ");
                            }
                            #endregion

                        }
						#endregion

                        #region last group
                        //This is only for the last group, which is not written into the cycle as the others
                        if (group.Key.Equals(groupedListLast.Key)) {
                            result.Replace(";", ".", result.Length - 4, 1);
                            sw.Write(result.ToString());
                        }
                        #endregion

                    }
                    #endregion

                }
                #endregion

            }
            catch (Exception ex) {
                throw new RDFModelException("Cannot serialize Turtle because: " + ex.Message);
            }
        }
        #endregion

    }

}