﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using RDFSharp.Model;
using RDFSharp.Store;

namespace RDFSharp.Query
{
    
    /// <summary>
    /// RDFRegexFilter represents a filter applying a regular expression on the values of a variable.
    /// </summary>
    public class RDFRegexFilter: RDFFilter {

        #region Properties
        /// <summary>
        /// Variable to be filtered
        /// </summary>
        public RDFVariable Variable { get; internal set; }

        /// <summary>
        /// Regular Expression to be filtered
        /// </summary>
        public Regex RegEx { get; internal set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build a filter on the given variable for the given regular expression 
        /// </summary>
        public RDFRegexFilter(RDFVariable variable, Regex regex) {
            if (variable != null) {
                if(regex != null) {
                    this.Variable = variable;
                    this.RegEx    = regex;
                    this.FilterID = RDFModelUtilities.CreateHash(this.ToString());   
                }
                else {
                    throw new RDFQueryException("Cannot create RDFRegexFilter because \"regex\" parameter is null.");
                }
            }
            else {
                throw new RDFQueryException("Cannot create RDFRegexFilter because \"variable\" parameter is null.");
            }
        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the filter 
        /// </summary>
        public override String ToString() {
            StringBuilder regexFlags = new StringBuilder();
            if (this.RegEx.Options.ToString().Contains(RegexOptions.IgnoreCase.ToString())) {
                regexFlags.Append("i");
            }
            if (this.RegEx.Options.ToString().Contains(RegexOptions.Singleline.ToString())) {
                regexFlags.Append("s");
            }
            if (this.RegEx.Options.ToString().Contains(RegexOptions.Multiline.ToString())) {
                regexFlags.Append("m");
            }
            if (this.RegEx.Options.ToString().Contains(RegexOptions.IgnorePatternWhitespace.ToString())) {
                regexFlags.Append("x");
            }
            if (regexFlags.ToString() != String.Empty) {
                return "FILTER ( REGEX(STR(" + this.Variable + "), \"" + this.RegEx + "\", \"" + regexFlags.ToString() + "\") )";
            }
            else {
                return "FILTER ( REGEX(STR(" + this.Variable + "), \"" + this.RegEx + "\") )";
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Applies the filter on the column corresponding to the variable in the given datarow
        /// </summary>
        internal override Boolean ApplyFilter(DataRow row, Boolean applyNegation) {
            Boolean keepRow = true;

            //Check is performed only if the row contains a column named like the filter's variable
            if (row.Table.Columns.Contains(this.Variable.ToString())) {
                String variableValue = row[this.Variable.ToString()].ToString();
            
                //Successfull match if the regular expression is satisfied by the variable
                keepRow     = this.RegEx.IsMatch(variableValue);

                //Apply the eventual negation
                if (applyNegation == true) {
                    keepRow = !keepRow;
                }
            }

            return keepRow;
        }
        #endregion

    }

}