﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using RDFSharp.Model;
using RDFSharp.Store;

namespace RDFSharp.Query
{

    /// <summary>
    /// RDFSelectQuery is a SPARQL query which gives a tabular response containing results for the given pattern groups and modifiers.
    /// </summary>
    public class RDFSelectQuery {

        #region Properties
        /// <summary>
        /// List of pattern groups carried by the query
        /// </summary>
        internal List<RDFPatternGroup> PatternGroups { get; set; }

        /// <summary>
        /// List of modifiers carried by the query
        /// </summary>
        internal List<RDFModifier> Modifiers { get; set; }

        /// <summary>
        /// Dictionary of pattern result tables
        /// </summary>
        internal Dictionary<RDFPatternGroup, List<DataTable>> PatternResultTables { get; set; }

        /// <summary>
        /// Dictionary of pattern group result tables
        /// </summary>
        internal Dictionary<RDFPatternGroup, DataTable> PatternGroupResultTables { get; set; }

        /// <summary>
        /// Checks if the query is a "SELECT *" query
        /// </summary>
        public Boolean IsStar {
            get {
                return (this.PatternGroups.TrueForAll(pg => pg.Variables.TrueForAll(v => v.IsResult == false) ||
                                                            pg.Variables.TrueForAll(v => v.IsResult == true)));
            }
        }

        /// <summary>
        /// Checks if the query is empty, so contains no pattern groups
        /// </summary>
        public Boolean IsEmpty {
            get { 
                return this.PatternGroups.Count == 0;
            }
        }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build an empty SELECT query
        /// </summary>
        public RDFSelectQuery() {
            this.PatternGroups            = new List<RDFPatternGroup>();
            this.Modifiers                = new List<RDFModifier>();
            this.PatternResultTables      = new Dictionary<RDFPatternGroup, List<DataTable>>();
            this.PatternGroupResultTables = new Dictionary<RDFPatternGroup, DataTable>();
        }

        /// <summary>
        /// List-based ctor to build a SELECT query from a list of pattern groups
        /// </summary>
        public RDFSelectQuery(List<RDFPatternGroup> patternGroups): this() {
            if (patternGroups != null) {
                patternGroups.ForEach(pg => {
                    this.AddPatternGroup(pg);
                });
            }
        }

        /// <summary>
        /// List-based ctor to build a SELECT query from a list of pattern groups and a list of modifiers
        /// </summary>
        public RDFSelectQuery(List<RDFPatternGroup> patternGroups, List<RDFModifier> modifiers): this(patternGroups) {
            if (modifiers != null) {
                modifiers.ForEach(m => {
                    this.AddModifier(m);
                });
            }
        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the SELECT query
        /// </summary>
        public override String ToString() {            
            StringBuilder query    = new StringBuilder();

            if (this.PatternGroups.Any()) {
                
                // SELECT
                query.Append("SELECT");

                // DISTINCT
                this.Modifiers.Where(mod => mod is RDFDistinctModifier).ToList<RDFModifier>().ForEach(dm => {
                    query.Append(" " + dm.ToString());
                });

                // VARIABLES
                if (this.IsStar  == true) {
                    query.Append(" *");
                }
                else {
                    List<RDFVariable> projVars = new List<RDFVariable>();
                    this.PatternGroups.ForEach(pg => {
                        var vEnum = pg.Variables.GetEnumerator();
                        while (vEnum.MoveNext()) {
                            if (vEnum.Current.IsResult == true) {
                                if (!projVars.Exists(v => v.Equals(vEnum.Current))) {
                                    projVars.Add(vEnum.Current);
                                }
                            }
                        }
                    });
                    projVars.ForEach(variable => {
                        query.Append(" " + variable.ToString());
                    });
                }

                // PATTERN GROUPS (pretty-printing of Unions)
                query.Append("\nWHERE\n{\n");
                Boolean printingUnion           = false;
                this.PatternGroups.ForEach(pg   => {

                    //Current pattern group is set as UNION with the next one
                    if (pg.JoinAsUnion         == true) {

                        //Current pattern group IS NOT the last of the query (so UNION keyword must be appended at last)
                        if(!pg.Equals(this.PatternGroups.Last())) {
                            //Begin a new Union block
                            if (printingUnion  == false) {
                                printingUnion   = true;
                                query.Append("\n  {");
                            }                            
                            query.Append(pg.ToString(2) + "    UNION");
                        }

                        //Current pattern group IS the last of the query (so UNION keyword must not be appended at last)
                        else {
                            //End the Union block
                            if (printingUnion  == true) {
                                printingUnion   = false;
                                query.Append(pg.ToString(2));
                                query.Append("  }\n");
                            }
                            else {
                                query.Append(pg.ToString());
                            }
                        }

                    }

                    //Current pattern group is set as INTERSECT with the next one
                    else {
                        //End the Union block
                        if (printingUnion    == true) {
                            printingUnion     = false;
                            query.Append(pg.ToString(2));
                            query.Append("  }\n");
                        }
                        else{
                            query.Append(pg.ToString());
                        }
                    }

                });
                query.Append("\n}");

                // ORDER BY
                if (this.Modifiers.Any(mod => mod is RDFOrderByModifier) == true) {
                    query.Append("\nORDER BY");
                    this.Modifiers.Where(mod => mod is RDFOrderByModifier).ToList<RDFModifier>().ForEach(om => {
                        query.Append(" " + om.ToString());
                    });
                }

                // LIMIT/OFFSET
                if (this.Modifiers.Any(mod => mod is RDFLimitModifier || mod is RDFOffsetModifier)) {
                    this.Modifiers.Where(mod => mod is RDFLimitModifier).ToList<RDFModifier>().ForEach(lim => {
                        query.Append("\n" + lim.ToString());
                    });
                    this.Modifiers.Where(mod => mod is RDFOffsetModifier).ToList<RDFModifier>().ForEach(off => {
                        query.Append("\n" + off.ToString());
                    });
                }

            }

            return query.ToString();
        }
        #endregion

        #region Methods
        /// <summary>
        /// Adds the given pattern group to the query
        /// </summary>
        public RDFSelectQuery AddPatternGroup(RDFPatternGroup patternGroup) {
            if (patternGroup != null) {
                if (!this.PatternGroups.Exists(pg => pg.PatternGroupName.Equals(patternGroup.PatternGroupName))) {
                    this.PatternGroups.Add(patternGroup);
                }
            }
            return this;
        }

        /// <summary>
        /// Adds the given modifier to the SELECT query
        /// </summary>
        public RDFSelectQuery AddModifier(RDFModifier modifier) {

            //Ensure to have only one distinct modifier in the query
            if (modifier is RDFDistinctModifier && this.Modifiers.Any(m => m is RDFDistinctModifier) == true) {
                return this;
            }
            //Ensure to have only one limit modifier in the query
            if (modifier is RDFLimitModifier    && this.Modifiers.Any(m => m is RDFLimitModifier)    == true) {
                return this;
            }
            //Ensure to have only one offset modifier in the query
            if (modifier is RDFOffsetModifier   && this.Modifiers.Any(m => m is RDFOffsetModifier)   == true) {
                return this;
            }

            if (!this.Modifiers.Exists(m => m.Equals(modifier))) {
                this.Modifiers.Add(modifier);
            }

            return this;
        }

        /// <summary>
        /// Applies the query to the given graph 
        /// </summary>
        public RDFSelectQueryResult ApplyToGraph(RDFGraph graph)  {
            if (graph != null) {
                RDFSelectQueryResult selectResult = new RDFSelectQueryResult(this.ToString());

                if (this.IsEmpty == false) {

                    //Step 0: Clean the query from previous execution results
                    this.PatternGroupResultTables.Clear();
                    this.PatternResultTables.Clear();

                    foreach (RDFPatternGroup patternGroup in this.PatternGroups) {

                        //Step 1: Get the intermediate result tables of the current pattern group
                        RDFQueryEngine.EvaluatePatterns(this, patternGroup, graph);

                        //Step 2: Get the result table of the current pattern group
                        RDFQueryEngine.CombinePatterns(this, patternGroup);

                        //Step 3: Apply the filters of the current pattern group to its result table
                        RDFQueryEngine.ApplyFilters(this, patternGroup);

                    }

                    //Step 4: Get the result table of the query
                    DataTable queryResultTable = RDFQueryEngine.CombineTables(this.PatternGroupResultTables.Values.ToList<DataTable>(), false);

                    //Step 5: Apply the modifiers of the query to the result table
                    selectResult.SelectResults = RDFQueryEngine.ApplyModifiers(this, queryResultTable);
                    selectResult.SelectResults.TableName = this.ToString();
                }

                return selectResult;
            }
            else {
                throw new RDFQueryException("Cannot execute SELECT query because given \"graph\" parameter is null.");
            }
        }

        /// <summary>
        /// Applies the query to the given store 
        /// </summary>
        public RDFSelectQueryResult ApplyToStore(RDFStore store) {
            if (store != null) {
                RDFSelectQueryResult selectResult = new RDFSelectQueryResult(this.ToString());

                if (this.IsEmpty == false) {

                    //Step 0: Clean the query from previous execution results
                    this.PatternGroupResultTables.Clear();
                    this.PatternResultTables.Clear();
                    
                    foreach (RDFPatternGroup patternGroup in this.PatternGroups) {

                        //Step 1: Get the intermediate result tables of the current pattern group
                        RDFQueryEngine.EvaluatePatterns(this, patternGroup, store);

                        //Step 2: Get the result table of the current pattern group
                        RDFQueryEngine.CombinePatterns(this, patternGroup);

                        //Step 3: Apply the filters of the current pattern group to its result table
                        RDFQueryEngine.ApplyFilters(this, patternGroup);

                    }

                    //Step 4: Get the result table of the query
                    DataTable queryResultTable = RDFQueryEngine.CombineTables(this.PatternGroupResultTables.Values.ToList<DataTable>(), false);

                    //Step 5: Apply the modifiers of the query to the result table
                    selectResult.SelectResults = RDFQueryEngine.ApplyModifiers(this, queryResultTable);
                    selectResult.SelectResults.TableName = this.ToString();
                }

                return selectResult;
            }
            else {
                throw new RDFQueryException("Cannot execute SELECT query because given \"store\" parameter is null.");
            }
        }

        /// <summary>
        /// Applies the query to the given federation
        /// </summary>
        public RDFSelectQueryResult ApplyToFederation(RDFFederation federation) {
            if (federation != null) {
                RDFSelectQueryResult selectResult        = new RDFSelectQueryResult(this.ToString());

                if (this.IsEmpty == false) {

                    //Step 1: Get the response of each federated store
                    List<DataTable> selectFedResults     = new List<DataTable>();
                    foreach (RDFStore store in federation.Stores.Values) {
                        selectFedResults.Add(this.ApplyToStore(store).SelectResults);
                    }

                    //Step 2: Combine the SELECT results
                    DataTable queryResultTable           = RDFQueryEngine.CombineTables(selectFedResults, true);

                    //Step 3: Apply the modifiers of the query to the result table
                    selectResult.SelectResults           = RDFQueryEngine.ApplyModifiers(this, queryResultTable);
                    selectResult.SelectResults.TableName = this.ToString();

                }

                return selectResult;
            }
            else {
                throw new RDFQueryException("Cannot execute SELECT query because given \"federation\" parameter is null.");
            }
        }
        #endregion

    }

}