﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Globalization;
using System.Linq;
using RDFSharp.Model;
using RDFSharp.Store;

namespace RDFSharp.Query
{

    /// <summary>
    /// RDFQueryUtilities is a collector of reusable utility methods for RDF query management
    /// </summary>
    internal static class RDFQueryUtilities {

        #region ADO.NET
        /// <summary>
        /// Adds a new column to the given table, avoiding duplicates 
        /// </summary>
        internal static void AddColumn(DataTable table, String columnName) {
            if (!table.Columns.Contains(columnName.Trim().ToUpperInvariant())) {
                table.Columns.Add(columnName.Trim().ToUpperInvariant(), Type.GetType("System.String"));
            }
        }

        /// <summary>
        /// Adds a new row to the given table 
        /// </summary>
        internal static void AddRow(DataTable table, Dictionary<String, String> bindings) {
            Boolean rowAdded     = false;
            DataRow resultRow    = table.NewRow();
            bindings.Keys.ToList<String>().ForEach(k => {
                if (table.Columns.Contains(k)) {
                    resultRow[k] = bindings[k];
                    rowAdded     = true;
                }
            });
            if (rowAdded == true) {
                table.Rows.Add(resultRow);
            }
        }
        #endregion

		#region OTHERS
        /// <summary>
        /// Parses the given string to return an instance of pattern member
        /// </summary>
        internal static RDFPatternMember ParseRDFPatternMember(String pMember) {

            #region URI
            Uri testUri;
            if (Uri.TryCreate(pMember, UriKind.Absolute, out testUri)) {
                return new RDFResource(pMember);
            }
            #endregion

            #region PLAIN LITERAL
            else if (!pMember.Contains("^^") || pMember.EndsWith("^^") || 
                RDFModelUtilities.GetUriFromString(pMember.Substring(pMember.LastIndexOf("^^") + 2)) == null) {
                RDFPlainLiteral pLit  = null;
                if (pMember.Contains("@")) {
                    if (!pMember.EndsWith("@")) {
                        Int32 lastAmp = pMember.LastIndexOf('@');
                        pLit          = new RDFPlainLiteral(pMember.Substring(0, lastAmp), pMember.Substring(lastAmp + 1));
                    }
                    else {
                        pLit          = new RDFPlainLiteral(pMember);
                    }
                }
                else {
                    pLit              = new RDFPlainLiteral(pMember);
                }
                return pLit;
            }
            #endregion

            #region TYPED LITERAL
            else {
                String tLitValue      = pMember.Substring(0, pMember.LastIndexOf("^^"));
                String tLitDatatype   = pMember.Substring(pMember.LastIndexOf("^^") + 2);
                RDFDatatype dt        = RDFModelUtilities.GetDatatypeFromString(tLitDatatype);
                RDFTypedLiteral tLit  = new RDFTypedLiteral(tLitValue, dt);
                return tLit;
            }
            #endregion

        }

        /// <summary>
        /// Compares the given pattern members, throwing a "Type Error" whenever the comparison operator detects sematically incompatible members;
        /// </summary>
        internal static Int32 CompareRDFPatternMembers(RDFPatternMember left, RDFPatternMember right) {

            #region CornerCase Comparisons
            if (ReferenceEquals(left, right)) {
                return 0;
            }
            if (left      == null) {
                if (right == null) {
                    return  0;
                }
                else {
                    return -1;
                }
            }
            else {
                if (right == null) {
                    return  1;
                }
            }
            #endregion

            #region Effective  Comparisons

            #region RESOURCE/CONTEXT
            if (left is RDFResource      || left is RDFContext) {

                //RESOURCE/CONTEXT VS RESOURCE/CONTEXT
                if (right is RDFResource || right is RDFContext) {
                    return left.ToString().CompareTo(right.ToString());
                }

                //RESOURCE/CONTEXT VS LITERAL
                else {

                    //RESOURCE/CONTEXT VS "XSD:ANYURI" TYPED LITERAL
                    if (right is RDFTypedLiteral && ((RDFTypedLiteral)right).Datatype.Equals(RDFDatatypeRegister.GetByPrefixAndDatatype(RDFVocabulary.XSD.PREFIX, "anyURI"))) {
                        return left.ToString().CompareTo(((RDFTypedLiteral)right).Value);
                    }

                    //RESOURCE/CONTEXT VS LITERAL
                    else{
                        return -1;
                    }

                }

            }
            #endregion

            #region PLAINLITERAL
            else if (left is RDFPlainLiteral) {

                //PLAIN LITERAL VS RESOURCE/CONTEXT
                if (right is RDFResource || right is RDFContext) {
                    return 1;
                }

                //PLAIN LITERAL VS PLAIN LITERAL
                else if (right is RDFPlainLiteral) {
                    return left.ToString().CompareTo(right.ToString());
                }

                //PLAIN LITERAL VS TYPED LITERAL
                else {

                    //PLAIN LITERAL VS "XSD:STRING" TYPED LITERAL
                    if (((RDFTypedLiteral)right).Datatype.Equals(RDFDatatypeRegister.GetByPrefixAndDatatype(RDFVocabulary.XSD.PREFIX, "string"))) {
                        return left.ToString().CompareTo(((RDFTypedLiteral)right).Value);
                    }

                    //PLAIN LITERAL VS TYPED LITERAL
                    else{
                        return -1;
                    }

                }
            }
            #endregion

            #region TYPEDLITERAL
            else {

                //TYPED LITERAL VS RESOURCE/CONTEXT
                if (right is RDFResource || right is RDFContext) {

                    //"XSD:ANYURI" TYPED LITERAL VS RESOURCE/CONTEXT
                    if (left is RDFTypedLiteral && ((RDFTypedLiteral)left).Datatype.Equals(RDFDatatypeRegister.GetByPrefixAndDatatype(RDFVocabulary.XSD.PREFIX, "anyURI"))) {
                        return ((RDFTypedLiteral)left).Value.CompareTo(right.ToString());
                    }

                    //TYPED LITERAL VS RESOURCE/CONTEXT
                    else {
                        return 1;
                    }
                }

                //TYPED LITERAL VS PLAIN LITERAL
                else if (right is RDFPlainLiteral) {

                    //"XSD:STRING" TYPED LITERAL VS PLAIN LITERAL
                    if (((RDFTypedLiteral)left).Datatype.Equals(RDFDatatypeRegister.GetByPrefixAndDatatype(RDFVocabulary.XSD.PREFIX, "string"))) {
                        return ((RDFTypedLiteral)left).Value.CompareTo(right.ToString());
                    }

                    //TYPED LITERAL VS PLAIN LITERAL
                    else {
                        return 1;
                    }

                }

                //TYPED LITERAL VS TYPED LITERAL
                else {

                    //SEMANTICALLY COMPATIBLE CATEGORY
                    if (((RDFTypedLiteral)left).Datatype.Category.Equals(((RDFTypedLiteral)right).Datatype.Category)) {
                        Int32 comparison = 0;
                        switch (((RDFTypedLiteral)left).Datatype.Category) {
                            case RDFModelEnums.RDFDatatypeCategory.Numeric:
                                Decimal leftValueDecimal    = Decimal.Parse(((RDFTypedLiteral)left).Value,  NumberStyles.Number, CultureInfo.InvariantCulture);
                                Decimal rightValueDecimal   = Decimal.Parse(((RDFTypedLiteral)right).Value, NumberStyles.Number, CultureInfo.InvariantCulture);
                                comparison                  = leftValueDecimal.CompareTo(rightValueDecimal);
                                break;
                            case RDFModelEnums.RDFDatatypeCategory.Boolean:
                                Boolean leftValueBoolean    = Boolean.Parse(((RDFTypedLiteral)left).Value);
                                Boolean rightValueBoolean   = Boolean.Parse(((RDFTypedLiteral)right).Value);
                                comparison                  = leftValueBoolean.CompareTo(rightValueBoolean);
                                break;
                            case RDFModelEnums.RDFDatatypeCategory.DateTime:
                                DateTime leftValueDateTime  = DateTime.Parse(((RDFTypedLiteral)left).Value,  CultureInfo.InvariantCulture, DateTimeStyles.AssumeUniversal);
                                DateTime rightValueDateTime = DateTime.Parse(((RDFTypedLiteral)right).Value, CultureInfo.InvariantCulture, DateTimeStyles.AssumeUniversal);
                                comparison                  = leftValueDateTime.CompareTo(rightValueDateTime);
                                break;
                            case RDFModelEnums.RDFDatatypeCategory.TimeSpan:
#if NET35
                                TimeSpan leftValueDuration  = TimeSpan.Parse(((RDFTypedLiteral)left).Value);
                                TimeSpan rightValueDuration = TimeSpan.Parse(((RDFTypedLiteral)right).Value);
#endif
#if NET40
                                TimeSpan leftValueDuration  = TimeSpan.Parse(((RDFTypedLiteral)left).Value,  CultureInfo.InvariantCulture);
                                TimeSpan rightValueDuration = TimeSpan.Parse(((RDFTypedLiteral)right).Value, CultureInfo.InvariantCulture);
#endif
                                comparison                  = leftValueDuration.CompareTo(rightValueDuration);
                                break;
                            case RDFModelEnums.RDFDatatypeCategory.String:
                                String leftValueString      = ((RDFTypedLiteral)left).Value;
                                String rightValueString     = ((RDFTypedLiteral)right).Value;
                                comparison                  = leftValueString.CompareTo(rightValueString);
                                break;
                        }
                        return comparison;
                    }

                    //SEMANTICALLY NOT COMPATIBLE CATEGORY
                    else {
                        throw new RDFQueryException("Type Error: Typed Literal cannot be compared to Typed Literal with different category.");
                    }

                }

            }
            #endregion

            #endregion

        }

        /// <summary>
        /// Gives a formatted string representation of the given pattern member
        /// </summary>
        internal static String PrintRDFPatternMember(RDFPatternMember patternMember) {
            if (patternMember is RDFVariable) {
                return patternMember.ToString();
            }
            else {
                if (patternMember is RDFResource || patternMember is RDFContext) {
                    if (patternMember is RDFResource && ((RDFResource)patternMember).IsBlank == true) {
                        return patternMember.ToString();
                    }
                    else{
                        return "<" + patternMember.ToString() + ">";
                    }
                }
                else {
                    if (patternMember is RDFPlainLiteral) {
                        if (((RDFPlainLiteral)patternMember).Language != String.Empty) {
                            return "\"" + ((RDFPlainLiteral)patternMember).Value + "\"@" + ((RDFPlainLiteral)patternMember).Language;
                        }
                        else {
                            return "\"" + ((RDFPlainLiteral)patternMember).Value + "\"";
                        }
                    }
                    else {
                        return "\"" + ((RDFTypedLiteral)patternMember).Value + "\"^^<" + ((RDFTypedLiteral)patternMember).Datatype + ">";
                    }
                }
            }
        }
        #endregion

    }

}