﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using FirebirdSql.Data.FirebirdClient;
using RDFSharp.Model;
using RDFSharp.Store;

namespace RDFSharp.Store
{

    /// <summary>
    /// RDFFirebirdStore represents a store backed on Firebird engine
    /// </summary>
    public class RDFFirebirdStore: RDFStore {

        #region Properties
        /// <summary>
        /// Connection to the Firebird database
        /// </summary>
        public FbConnection Connection { get; internal set; }

        /// <summary>
        /// Dictionary of commands to manipulate the Firebird database
        /// </summary>
        internal Dictionary<String, FbCommand> Commands { get; set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build a local Firebird store instance ("LOCALHOST") at the given path
        /// </summary>
        public RDFFirebirdStore(String dbPath): this("LOCALHOST", dbPath) { }

        /// <summary>
        /// Default-ctor to build a Firebird store instance at the given path
        /// </summary>
        public RDFFirebirdStore(String firebirdInstance, String dbPath): base() {
			if(firebirdInstance != null && firebirdInstance.Trim() != String.Empty){
				firebirdInstance    = firebirdInstance.Trim();
				if(RDFModelUtilities.CheckLocalPath(dbPath) == true){
					dbPath          = dbPath.Trim();

					//Initialize store structures to Firebird values
					this.StoreType  = "FIREBIRD";
                    this.Connection = new FbConnection(@"DataSource=" + firebirdInstance + ";Database=" + dbPath + ";User=SYSDBA;Password=masterkey;ServerType=0;Dialect=3;Charset=NONE;");
					this.Commands   = new Dictionary<String, FbCommand>();
                    this.StoreID    = RDFModelUtilities.CreateHash(this.ToString());

					#region PrepareCommands
					//SELECT
					this.Commands.Add("SELECT",   new FbCommand("SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples", this.Connection));
					//INSERT
                    this.Commands.Add("INSERT",   new FbCommand("UPDATE OR INSERT INTO Quadruples(QuadrupleID, TripleFlavor, Context, ContextID, Subject, SubjectID, Predicate, PredicateID, Object, ObjectID) VALUES (@QID, @TFV, @CTX, @CTXID, @SUBJ, @SUBJID, @PRED, @PREDID, @OBJ, @OBJID) MATCHING (QuadrupleID)", this.Connection));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("QID",    FbDbType.BigInt));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("TFV",    FbDbType.Integer));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("CTX",    FbDbType.VarChar, 1000));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("CTXID",  FbDbType.BigInt));
                        this.Commands["INSERT"].Parameters.Add(new FbParameter("SUBJ",   FbDbType.VarChar, 1000));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("SUBJID", FbDbType.BigInt));
                        this.Commands["INSERT"].Parameters.Add(new FbParameter("PRED",   FbDbType.VarChar, 1000));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("PREDID", FbDbType.BigInt));
                        this.Commands["INSERT"].Parameters.Add(new FbParameter("OBJ",    FbDbType.VarChar, 2000));
						this.Commands["INSERT"].Parameters.Add(new FbParameter("OBJID",  FbDbType.BigInt));
					//DELETE
					this.Commands.Add("DELETE",   new FbCommand("DELETE FROM Quadruples WHERE QuadrupleID = @QID", this.Connection));
						this.Commands["DELETE"].Parameters.Add(new FbParameter("QID",    FbDbType.BigInt));
					#endregion

					//If the given file does not exist, clone the Firebird template database
					if(!File.Exists(dbPath)) {

						#region CloneTemplate
						try {
							Assembly firebird        = Assembly.GetExecutingAssembly();
                            using (var templateDB    = firebird.GetManifestResourceStream("RDFSharp.Store.Template.RDFFirebirdTemplate.fdb")) {
								using (var destineDB = new FileStream(dbPath, FileMode.Create, FileAccess.ReadWrite)) {

#if NET40
                                    templateDB.CopyTo(destineDB);
#endif
#if NET35
                                    Byte[] buffer    = new Byte[4096];
                                    Int32 count      = 0;
                                    while ((count    = templateDB.Read(buffer, 0, buffer.Length)) != 0) {
                                        destineDB.Write(buffer, 0, count);
                                    }
#endif

								}
							}
						}
						catch (Exception ex) {
							throw new RDFStoreException("Cannot create Firebird store because: " + ex.Message);
						}
						#endregion

					}

				}
				else {
					throw new RDFStoreException("Cannot connect to Firebird store because \"dbPath\" parameter is null or does not indicate a local file");
				}
			}
			else {
                throw new RDFStoreException("Cannot connect to Firebird store because \"firebirdInstance\" parameter is null or empty");
            }
        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the Firebird store 
        /// </summary>
        public override String ToString() {
            return base.ToString() + "|CONNECTION=" + this.Connection.ConnectionString;
        }
        #endregion

        #region Methods

        #region Add
        /// <summary>
        /// Merges the given graph into the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore MergeGraph(RDFGraph graph) {
            if (graph != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["INSERT"].Transaction = this.Commands["INSERT"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["INSERT"].Prepare();
                    //Iterate graph triples
                    RDFContext graphCtx    = new RDFContext(graph.Context);
                    RDFQuadruple quadruple = null;
                    var enumGraph          = graph.TriplesEnumerator;
                    while (enumGraph.MoveNext()) {
                        if (enumGraph.Current.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO) {
                            quadruple      = new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFResource)enumGraph.Current.Object);
                        }
                        else {
                            quadruple      = new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFLiteral)enumGraph.Current.Object);
                        }
                        //Valorize parameters
                        this.Commands["INSERT"].Parameters["QID"].Value    = quadruple.QuadrupleID;
                        this.Commands["INSERT"].Parameters["TFV"].Value    = quadruple.TripleFlavor;
                        this.Commands["INSERT"].Parameters["CTX"].Value    = quadruple.Context.ToString();
                        this.Commands["INSERT"].Parameters["CTXID"].Value  = quadruple.Context.PatternMemberID;
                        this.Commands["INSERT"].Parameters["SUBJ"].Value   = quadruple.Subject.ToString();
                        this.Commands["INSERT"].Parameters["SUBJID"].Value = quadruple.Subject.PatternMemberID;
                        this.Commands["INSERT"].Parameters["PRED"].Value   = quadruple.Predicate.ToString();
                        this.Commands["INSERT"].Parameters["PREDID"].Value = quadruple.Predicate.PatternMemberID;
                        this.Commands["INSERT"].Parameters["OBJ"].Value    = quadruple.Object.ToString();
                        this.Commands["INSERT"].Parameters["OBJID"].Value  = quadruple.Object.PatternMemberID;
                        //Execute command
                        this.Commands["INSERT"].ExecuteNonQuery();
                    }
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot insert data into Firebird store because: " + ex.Message);
                }
            }
            return this;
        }

        /// <summary>
        /// Adds the given quadruple to the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore AddQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["INSERT"].Transaction = this.Commands["INSERT"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["INSERT"].Prepare();
                    //Valorize parameters
                    this.Commands["INSERT"].Parameters["QID"].Value    = quadruple.QuadrupleID;
                    this.Commands["INSERT"].Parameters["TFV"].Value    = quadruple.TripleFlavor;
                    this.Commands["INSERT"].Parameters["CTX"].Value    = quadruple.Context.ToString();
                    this.Commands["INSERT"].Parameters["CTXID"].Value  = quadruple.Context.PatternMemberID;
                    this.Commands["INSERT"].Parameters["SUBJ"].Value   = quadruple.Subject.ToString();
                    this.Commands["INSERT"].Parameters["SUBJID"].Value = quadruple.Subject.PatternMemberID;
                    this.Commands["INSERT"].Parameters["PRED"].Value   = quadruple.Predicate.ToString();
                    this.Commands["INSERT"].Parameters["PREDID"].Value = quadruple.Predicate.PatternMemberID;
                    this.Commands["INSERT"].Parameters["OBJ"].Value    = quadruple.Object.ToString();
                    this.Commands["INSERT"].Parameters["OBJID"].Value  = quadruple.Object.PatternMemberID;
                    //Execute command
                    this.Commands["INSERT"].ExecuteNonQuery();
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot insert data into Firebird store because: " + ex.Message);
                }
            }
            return this;
        }
        #endregion

        #region Remove
        /// <summary>
        /// Removes the given quadruples from the store
        /// </summary>
        public override RDFStore RemoveQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["DELETE"].Transaction = this.Commands["DELETE"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["DELETE"].Prepare();
                    //Valorize parameters
                    this.Commands["DELETE"].Parameters["QID"].Value = quadruple.QuadrupleID;
                    //Execute command
                    this.Commands["DELETE"].ExecuteNonQuery();
                    //Close transaction
                    this.Commands["DELETE"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["DELETE"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot delete data from Firebird store because: " + ex.Message);
                }
            }
            return this;
        }

        /// <summary>
        /// Clears the quadruples of the store
        /// </summary>
        public override RDFStore ClearQuadruples() {
            try {
                //Modify command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples";
                this.Commands["DELETE"].Parameters.Clear();
                //Open connection
                this.Connection.Open();
                //Open transaction
                this.Commands["DELETE"].Transaction = this.Commands["DELETE"].Connection.BeginTransaction();
                //Execute command
                this.Commands["DELETE"].ExecuteNonQuery();
                //Close transaction
                this.Commands["DELETE"].Transaction.Commit();
                //Close connection
                this.Connection.Close();
                //Restore command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples WHERE QuadrupleID = @QID";
                this.Commands["DELETE"].Parameters.Add(new FbParameter("QID", FbDbType.BigInt)); 
            }
            catch (Exception ex) {
                //Close transaction
                this.Commands["DELETE"].Transaction.Rollback();
                //Close connection
                this.Connection.Close();
                //Restore command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples WHERE QuadrupleID = @QID";
                this.Commands["DELETE"].Parameters.Add(new FbParameter("QID", FbDbType.BigInt));
                //Propagate exception
                throw new RDFStoreException("Cannot delete data from Firebird store because: " + ex.Message);
            }
            return this;
        }
        #endregion

        #region Select
        /// <summary>
        /// Gets a store containing all quadruples
        /// </summary>
        public override RDFStore SelectAllQuadruples() {
            RDFMemoryStore result                         = new RDFMemoryStore();
            try {
                //Open connection
                this.Connection.Open();
                //Execute command
                using (FbDataReader fetchedQuadruples     = this.Commands["SELECT"].ExecuteReader()) {
                    if (fetchedQuadruples.HasRows) {
                        //Iterate fetched quadruples
                        while (fetchedQuadruples.Read()) {
                            //Add the fetched quadruple to the result
                            result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                        }
                    }
                }
                //Close connection
                this.Connection.Close();
            }
            catch (Exception ex) {
                //Close connection
                this.Connection.Close();
                //Propagate exception
                throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified context
        /// </summary>
        public override RDFStore SelectQuadruplesByContext(RDFContext contextResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (contextResource  != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE ContextID = @CTXID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new FbParameter("CTXID", FbDbType.BigInt));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["CTXID"].Value = contextResource.PatternMemberID;
                    //Execute command
                    using (FbDataReader fetchedQuadruples             = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified subject
        /// </summary>
        public override RDFStore SelectQuadruplesBySubject(RDFResource subjectResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (subjectResource  != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText               += " WHERE SubjectID = @SUBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new FbParameter("SUBJID", FbDbType.BigInt));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["SUBJID"].Value = subjectResource.PatternMemberID;
                    //Execute command
                    using (FbDataReader fetchedQuadruples              = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified predicate
        /// </summary>
        public override RDFStore SelectQuadruplesByPredicate(RDFResource predicateResource) {
            RDFMemoryStore result  = new RDFMemoryStore();
            if (predicateResource != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText               += " WHERE PredicateID = @PREDID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new FbParameter("PREDID", FbDbType.BigInt));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["PREDID"].Value = predicateResource.PatternMemberID;
                    //Execute command
                    using (FbDataReader fetchedQuadruples              = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified object
        /// </summary>
        public override RDFStore SelectQuadruplesByObject(RDFResource objectResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (objectResource   != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE TripleFlavor = 1 AND ObjectID = @OBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new FbParameter("OBJID", FbDbType.BigInt));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["OBJID"].Value = objectResource.PatternMemberID;
                    //Execute command
                    using (FbDataReader fetchedQuadruples             = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified literal
        /// </summary>
        public override RDFStore SelectQuadruplesByLiteral(RDFLiteral objectLiteral) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (objectLiteral    != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE TripleFlavor = 2 AND ObjectID = @OBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new FbParameter("OBJID", FbDbType.BigInt));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["OBJID"].Value = objectLiteral.PatternMemberID;
                    //Execute command
                    using (FbDataReader fetchedQuadruples             = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from Firebird store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Parses the current quadruple of the Firebird reader 
        /// </summary>
        internal RDFQuadruple ParseQuadruple(FbDataReader fetchedQuadruples) {
            RDFContext qContext           = new RDFContext(fetchedQuadruples["Context"].ToString());
            RDFResource qSubject          = new RDFResource(fetchedQuadruples["Subject"].ToString());
            RDFResource qPredicate        = new RDFResource(fetchedQuadruples["Predicate"].ToString());
            //SPO-flavour quadruple
            if (fetchedQuadruples["TripleFlavor"].ToString() == "1") {
                RDFResource qObject       = new RDFResource(fetchedQuadruples["Object"].ToString());
                return new RDFQuadruple(qContext, qSubject, qPredicate, qObject);
            }
            //SPL-flavour quadruple
            else {
                String literal            = fetchedQuadruples["Object"].ToString();
                //PlainLiteral
                if (!literal.Contains("^^") || literal.EndsWith("^^") || RDFModelUtilities.GetUriFromString(literal.Substring(literal.LastIndexOf("^^") + 2)) == null) {
                    RDFPlainLiteral pLit  = null;
                    if (literal.Contains("@")) {
                        if (!literal.EndsWith("@")) {
                            Int32 lastAmp = literal.LastIndexOf('@');
                            pLit          = new RDFPlainLiteral(literal.Substring(0, lastAmp), literal.Substring(lastAmp + 1));
                        }
                        else {
                            pLit          = new RDFPlainLiteral(literal);
                        }
                    }
                    else {
                        pLit              = new RDFPlainLiteral(literal);
                    }
                    return new RDFQuadruple(qContext, qSubject, qPredicate, pLit);
                }
                //TypedLiteral
                else {
                    String tLitValue      = literal.Substring(0, literal.LastIndexOf("^^"));
                    String tLitDatatype   = literal.Substring(literal.LastIndexOf("^^") + 2);
                    RDFDatatype dt        = RDFModelUtilities.GetDatatypeFromString(tLitDatatype);
                    RDFTypedLiteral tLit  = new RDFTypedLiteral(tLitValue, dt);
                    return new RDFQuadruple(qContext, qSubject, qPredicate, tLit);
                }
            }
        }
        #endregion

        #endregion

    }

}