﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using RDFSharp.Model;

namespace RDFSharp.Store
{
    
    /// <summary>
    /// RDFMemoryStore represents an in-memory RDF store engine.
    /// </summary>
    public class RDFMemoryStore: RDFStore {

        #region Properties
        /// <summary>
        /// Count of the store's quadruples
        /// </summary>
        public Int64 QuadruplesCount {
            get { return this.Quadruples.Count; }
        }

        /// <summary>
        /// Gets the enumerator on the store's quadruples for iteration
        /// </summary>
        public IEnumerator<RDFQuadruple> QuadruplesEnumerator {
            get { return this.Quadruples.Values.GetEnumerator(); }
        }

        /// <summary>
        /// List of quadruples embedded into the store
        /// </summary>
        internal Dictionary<Int64, RDFQuadruple> Quadruples { get; set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build an empty memory store
        /// </summary>
        public RDFMemoryStore() {
            this.StoreType  = "MEMORY";
            this.StoreID    = RDFModelUtilities.CreateHash(this.ToString());
            this.Quadruples = new Dictionary<Int64, RDFQuadruple>();
        }

        /// <summary>
        /// List-based ctor to build a memory store with the given list of quadruples
        /// </summary>
        public RDFMemoryStore(List<RDFQuadruple> quadruples): this() {
            if (quadruples != null) {
                quadruples.ForEach(q => { 
                    this.AddQuadruple(q); 
                });
            }
        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the Memory store 
        /// </summary>
        public override String ToString() {
            return base.ToString() + "|ID=" + Guid.NewGuid().ToString();
        }

        /// <summary>
        /// Performs the equality comparison between two memory stores
        /// </summary>
        public Boolean Equals(RDFMemoryStore other) {
            if (ReferenceEquals(this, other)) {
                return true;
            }
            if (this      == null) {
                if (other == null) {
                    return true;
                }
                else {
                    return false;
                }
            }
            else {
                if (other == null) {
                    return false;
                }
                else {
                    //Effective Comparison
                    if (this.QuadruplesCount != other.QuadruplesCount) {
                        return false;
                    }
                    var qEnum = this.QuadruplesEnumerator;
                    while (qEnum.MoveNext()) {
                        if (!other.Quadruples.ContainsKey(qEnum.Current.QuadrupleID)) {
                            return false;
                        }
                    }
                    return true;
                }
            }
        }
        #endregion

        #region Methods

        #region Add
        /// <summary>
        /// Merges the given graph into the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore MergeGraph(RDFGraph graph) {
            if (graph != null) {
                RDFContext graphCtx    = new RDFContext(graph.Context);
                var enumGraph          = graph.TriplesEnumerator;                
                while (enumGraph.MoveNext()) {
                    if (enumGraph.Current.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO) {
                        this.AddQuadruple(new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFResource)enumGraph.Current.Object));
                    }
                    else {
                        this.AddQuadruple(new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFLiteral)enumGraph.Current.Object));
                    }
                }
            }
            return this;
        }

        /// <summary>
        /// Adds the given quadruple to the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore AddQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                if (!this.Quadruples.ContainsKey(quadruple.QuadrupleID)) {
                    this.Quadruples.Add(quadruple.QuadrupleID, quadruple);
                }
            }
            return this;
        }
        #endregion

        #region Remove
        /// <summary>
        /// Removes the given quadruples from the store
        /// </summary>
        public override RDFStore RemoveQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                if (this.Quadruples.ContainsKey(quadruple.QuadrupleID)) {
                    this.Quadruples.Remove(quadruple.QuadrupleID);
                }
            }
            return this;
        }

        /// <summary>
        /// Clears the quadruples of the store
        /// </summary>
        public override RDFStore ClearQuadruples() {
            this.Quadruples.Clear();
            return this;
        }
        #endregion

        #region Select
        /// <summary>
        /// Gets a store containing all quadruples
        /// </summary>
        public override RDFStore SelectAllQuadruples() {
            return this;
        }

        /// <summary>
        /// Gets a store containing quadruples with the specified context
        /// </summary>
        public override RDFStore SelectQuadruplesByContext(RDFContext contextResource) {
            List<RDFQuadruple> matchingQuadruples = new List<RDFQuadruple>();
            if (contextResource != null) {
                matchingQuadruples.AddRange(((from t in this.Quadruples.Values.ToList<RDFQuadruple>()
                                              where t.Context.Equals(contextResource)
                                              select t).ToList<RDFQuadruple>()));
            }
            return new RDFMemoryStore(matchingQuadruples);
        }

        /// <summary>
        /// Gets a store containing quadruples with the specified subject 
        /// </summary>
        public override RDFStore SelectQuadruplesBySubject(RDFResource subjectResource) {
            List<RDFQuadruple> matchingQuadruples = new List<RDFQuadruple>();
            if (subjectResource != null) {
                matchingQuadruples.AddRange(((from t in this.Quadruples.Values.ToList<RDFQuadruple>()
                                              where t.Subject.Equals(subjectResource)
                                              select t).ToList<RDFQuadruple>()));
            }
            return new RDFMemoryStore(matchingQuadruples);
        }

        /// <summary>
        /// Gets a store containing quadruples with the specified predicate
        /// </summary>
        public override RDFStore SelectQuadruplesByPredicate(RDFResource predicateResource) {
            List<RDFQuadruple> matchingQuadruples = new List<RDFQuadruple>();
            if (predicateResource != null) {
                matchingQuadruples.AddRange(((from t in this.Quadruples.Values.ToList<RDFQuadruple>()
                                              where t.Predicate.Equals(predicateResource)
                                              select t).ToList<RDFQuadruple>()));
            }
            return new RDFMemoryStore(matchingQuadruples);
        }

        /// <summary>
        /// Gets a store containing quadruples with the specified object 
        /// </summary>
        public override RDFStore SelectQuadruplesByObject(RDFResource objectResource) {
            List<RDFQuadruple> matchingQuadruples = new List<RDFQuadruple>();
            if (objectResource != null) {
                matchingQuadruples.AddRange(((from t in this.Quadruples.Values.ToList<RDFQuadruple>()
                                              where t.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO && t.Object.Equals(objectResource)
                                              select t).ToList<RDFQuadruple>()));
            }
            return new RDFMemoryStore(matchingQuadruples);
        }

        /// <summary>
        /// Gets a store containing quadruples with the specified literal 
        /// </summary>
        public override RDFStore SelectQuadruplesByLiteral(RDFLiteral objectLiteral)  {
            List<RDFQuadruple> matchingQuadruples = new List<RDFQuadruple>();
            if (objectLiteral != null) {
                matchingQuadruples.AddRange(((from t in this.Quadruples.Values.ToList<RDFQuadruple>()
                                              where t.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPL && t.Object.Equals(objectLiteral)
                                              select t).ToList<RDFQuadruple>()));
            }
            return new RDFMemoryStore(matchingQuadruples);
        }
        #endregion

        #region Set
        /// <summary>
        /// Builds a new intersection store from this store and a given one
        /// </summary>
        public RDFStore IntersectWith(RDFStore store) {
            RDFStore result   = new RDFMemoryStore();
            if (store != null) {
                if (store is RDFMemoryStore) {
                    var qEnum = ((RDFMemoryStore)store).QuadruplesEnumerator;
                    while (qEnum.MoveNext()) {
                        if (this.Quadruples.ContainsKey(qEnum.Current.QuadrupleID)) {
                            result.AddQuadruple(qEnum.Current);
                        }
                    }
                }
                else {
                    result    = this.IntersectWith(store.SelectAllQuadruples());
                }
            }
            return result;
        }

        /// <summary>
        /// Builds a new union store from this store and a given one
        /// </summary>
        public RDFStore UnionWith(RDFStore store) {
            RDFStore result   = new RDFMemoryStore(this.Quadruples.Values.ToList<RDFQuadruple>());
            if (store != null) {
                if (store is RDFMemoryStore) {
                    var qEnum = ((RDFMemoryStore)store).QuadruplesEnumerator;
                    while (qEnum.MoveNext()) {
                        result.AddQuadruple(qEnum.Current);
                    }
                }
                else {
                    result    = this.UnionWith(store.SelectAllQuadruples());
                }
            }
            return result;
        }

        /// <summary>
        /// Builds a new difference store from this store and a given one
        /// </summary>
        public RDFStore DifferenceWith(RDFStore store) {
            RDFStore result   = new RDFMemoryStore(this.Quadruples.Values.ToList<RDFQuadruple>());
            if (store != null) {
                if (store is RDFMemoryStore) {
                    var qEnum = this.QuadruplesEnumerator;
                    while (qEnum.MoveNext()) {
                        if (((RDFMemoryStore)store).Quadruples.ContainsKey(qEnum.Current.QuadrupleID)) {
                            result.RemoveQuadruple(qEnum.Current);
                        }
                    }
                }
                else {
                    result    = this.DifferenceWith(store.SelectAllQuadruples());
                }
            }
            return result;
        }
        #endregion

        #endregion

    }

}