﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Data;
using MySql.Data.MySqlClient;
using RDFSharp.Model;
using RDFSharp.Store;

namespace RDFSharp.Store
{

    /// <summary>
    /// RDFMySQLStore represents a store backed on MySQL engine
    /// </summary>
    public class RDFMySQLStore: RDFStore {

        #region Properties
        /// <summary>
        /// Connection to the MySQL database
        /// </summary>
        public MySqlConnection Connection { get; internal set; }

        /// <summary>
        /// Dictionary of commands to manipulate the MySQL database
        /// </summary>
        internal Dictionary<String, MySqlCommand> Commands { get; set; }
        #endregion

        #region Ctors
        /// <summary>
        /// Default-ctor to build a local MySQL store instance ("LOCALHOST")
        /// </summary>
        public RDFMySQLStore(String mysqlDatabase, String mysqlUser, String mysqlPassword): this("LOCALHOST", mysqlDatabase, mysqlUser, mysqlPassword) { }

        /// <summary>
        /// Default-ctor to build a MySQL store instance
        /// </summary>
        public RDFMySQLStore(String mysqlInstance, String mysqlDatabase, String mysqlUser, String mysqlPassword): base() {
            if(mysqlInstance             != null && mysqlInstance.Trim() != String.Empty){
                mysqlInstance             = mysqlInstance.Trim();
                if(mysqlDatabase         != null && mysqlDatabase.Trim() != String.Empty) {
                    mysqlDatabase         = mysqlDatabase.Trim();
                    if(mysqlUser         != null && mysqlUser.Trim()     != String.Empty) {
                        mysqlUser         = mysqlUser.Trim();
                        if(mysqlPassword != null) {

                            //Initialize store structures to MySQL values
                            this.StoreType    = "MYSQL";
                            this.Connection   = new MySqlConnection(@"Server=" + mysqlInstance + ";Database=" + mysqlDatabase + ";Uid=" + mysqlUser + ";Pwd=" + mysqlPassword + ";Port=3306");
                            this.Commands     = new Dictionary<String, MySqlCommand>();
                            this.StoreID      = RDFModelUtilities.CreateHash(this.ToString());                       

                            #region PrepareCommands
                            //SELECT
                            this.Commands.Add("SELECT",   new MySqlCommand());
                                this.Commands["SELECT"].Connection       = this.Connection;
                                this.Commands["SELECT"].CommandType      = CommandType.Text;
                                this.Commands["SELECT"].CommandText      = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                            //INSERT
                            this.Commands.Add("INSERT",   new MySqlCommand());
                                this.Commands["INSERT"].Connection       = this.Connection;
                                this.Commands["INSERT"].CommandType      = CommandType.Text;
                                this.Commands["INSERT"].CommandText      = "INSERT IGNORE INTO Quadruples(QuadrupleID, TripleFlavor, Context, ContextID, Subject, SubjectID, Predicate, PredicateID, Object, ObjectID) VALUES (@QID, @TFV, @CTX, @CTXID, @SUBJ, @SUBJID, @PRED, @PREDID, @OBJ, @OBJID)";
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("QID",    MySqlDbType.Int64));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("TFV",    MySqlDbType.Int32));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("CTX",    MySqlDbType.VarChar, 1000));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("CTXID",  MySqlDbType.Int64));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("SUBJ",   MySqlDbType.VarChar, 1000));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("SUBJID", MySqlDbType.Int64));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("PRED",   MySqlDbType.VarChar, 1000));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("PREDID", MySqlDbType.Int64));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("OBJ",    MySqlDbType.VarChar, 2000));
                                this.Commands["INSERT"].Parameters.Add(new MySqlParameter("OBJID",  MySqlDbType.Int64));
                            //DELETE
                            this.Commands.Add("DELETE",   new MySqlCommand());
                                this.Commands["DELETE"].CommandType      = CommandType.Text;
                                this.Commands["DELETE"].Connection       = this.Connection;
                                this.Commands["DELETE"].CommandText      = "DELETE FROM Quadruples WHERE QuadrupleID = @QID";
                                this.Commands["DELETE"].Parameters.Add(new MySqlParameter("QID",    MySqlDbType.Int64));
                            //DIAGNOSTICS
                            this.Commands.Add("DIAGNOSTICS", new MySqlCommand());
                                this.Commands["DIAGNOSTICS"].Connection  = this.Connection;
                                this.Commands["DIAGNOSTICS"].CommandType = CommandType.Text;
                                this.Commands["DIAGNOSTICS"].CommandText = "SELECT COUNT(*) FROM information_schema.tables WHERE table_schema = '" + mysqlDatabase + "' AND table_name = 'Quadruples'";
                            //PREPARE
                            this.Commands.Add("PREPARE", new MySqlCommand());
                                this.Commands["PREPARE"].Connection      = this.Connection;
                                this.Commands["PREPARE"].CommandType     = CommandType.Text;
                                this.Commands["PREPARE"].CommandText     = "CREATE TABLE Quadruples(QuadrupleID BIGINT NOT NULL PRIMARY KEY, TripleFlavor INT NOT NULL, Context VARCHAR(1000) NOT NULL, ContextID BIGINT NOT NULL, Subject VARCHAR(1000) NOT NULL, SubjectID BIGINT NOT NULL, Predicate VARCHAR(1000) NOT NULL, PredicateID BIGINT NOT NULL, Object VARCHAR(2000) NOT NULL, ObjectID BIGINT NOT NULL) ENGINE=InnoDB;";
                                this.Commands["PREPARE"].CommandText    += "ALTER TABLE Quadruples ADD INDEX IDX_TripleFlavor(TripleFlavor);";
                                this.Commands["PREPARE"].CommandText    += "ALTER TABLE Quadruples ADD INDEX IDX_ContextID(ContextID);";
                                this.Commands["PREPARE"].CommandText    += "ALTER TABLE Quadruples ADD INDEX IDX_SubjectID(SubjectID);";
                                this.Commands["PREPARE"].CommandText    += "ALTER TABLE Quadruples ADD INDEX IDX_PredicateID(PredicateID);";
                                this.Commands["PREPARE"].CommandText    += "ALTER TABLE Quadruples ADD INDEX IDX_ObjectID(ObjectID);";
                            #endregion

                            #region Diagnostics
                            RDFStoreEnums.RDFStoreSQLErrors connCheck = this.MySQLDiagnostics();
                            switch(connCheck) {
                                case RDFStoreEnums.RDFStoreSQLErrors.NoErrors:
                                    break;
                                case RDFStoreEnums.RDFStoreSQLErrors.QuadruplesTableNotFound:
                                    this.PrepareDatabase();
                                    break;                
                                case RDFStoreEnums.RDFStoreSQLErrors.InvalidDataSource:
                                    throw new RDFStoreException("Cannot connect to MySQL store because unable to find the given MySQL instance or the given MySQL database.");
                            }
                            #endregion

                        }
                        else {
                            throw new RDFStoreException("Cannot connect to MySQL store because \"mysqlPassword\" parameter is null");
                        }
                    }
                    else {
                        throw new RDFStoreException("Cannot connect to MySQL store because \"mysqlUser\" parameter is null or empty");
                    }
                }
                else {
                    throw new RDFStoreException("Cannot connect to MySQL store because \"mysqlDatabase\" parameter is null or empty");
                }
            }
            else {
                throw new RDFStoreException("Cannot connect to MySQL store because \"mysqlInstance\" parameter is null or empty");
            }
        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the MySQL store 
        /// </summary>
        public override String ToString() {
            return base.ToString() + "|CONNECTION=" + this.Connection.ConnectionString;
        }
        #endregion

        #region Methods

        #region Add
        /// <summary>
        /// Merges the given graph into the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore MergeGraph(RDFGraph graph) {
            if (graph != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["INSERT"].Transaction = this.Commands["INSERT"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["INSERT"].Prepare();
                    //Iterate graph triples
                    RDFContext graphCtx     = new RDFContext(graph.Context);
                    RDFQuadruple quadruple  = null;
                    var enumGraph           = graph.TriplesEnumerator;                    
                    while (enumGraph.MoveNext()) {
                        if (enumGraph.Current.TripleFlavor == RDFModelEnums.RDFTripleFlavors.SPO) {
                            quadruple       = new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFResource)enumGraph.Current.Object);
                        }
                        else {
                            quadruple       = new RDFQuadruple(graphCtx, (RDFResource)enumGraph.Current.Subject, (RDFResource)enumGraph.Current.Predicate, (RDFLiteral)enumGraph.Current.Object);
                        }
                        //Valorize parameters
                        this.Commands["INSERT"].Parameters["QID"].Value    = quadruple.QuadrupleID;
                        this.Commands["INSERT"].Parameters["TFV"].Value    = quadruple.TripleFlavor;
                        this.Commands["INSERT"].Parameters["CTX"].Value    = quadruple.Context.ToString();
                        this.Commands["INSERT"].Parameters["CTXID"].Value  = quadruple.Context.PatternMemberID;
                        this.Commands["INSERT"].Parameters["SUBJ"].Value   = quadruple.Subject.ToString();
                        this.Commands["INSERT"].Parameters["SUBJID"].Value = quadruple.Subject.PatternMemberID;
                        this.Commands["INSERT"].Parameters["PRED"].Value   = quadruple.Predicate.ToString();
                        this.Commands["INSERT"].Parameters["PREDID"].Value = quadruple.Predicate.PatternMemberID;
                        this.Commands["INSERT"].Parameters["OBJ"].Value    = quadruple.Object.ToString();
                        this.Commands["INSERT"].Parameters["OBJID"].Value  = quadruple.Object.PatternMemberID;
                        //Execute command
                        this.Commands["INSERT"].ExecuteNonQuery();
                    }
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot insert data into MySQL store because: " + ex.Message);
                }
            }
            return this;
        }

        /// <summary>
        /// Adds the given quadruple to the store, avoiding duplicate insertions
        /// </summary>
        public override RDFStore AddQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["INSERT"].Transaction = this.Commands["INSERT"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["INSERT"].Prepare();
                    //Valorize parameters
                    this.Commands["INSERT"].Parameters["QID"].Value    = quadruple.QuadrupleID;
                    this.Commands["INSERT"].Parameters["TFV"].Value    = quadruple.TripleFlavor;
                    this.Commands["INSERT"].Parameters["CTX"].Value    = quadruple.Context.ToString();
                    this.Commands["INSERT"].Parameters["CTXID"].Value  = quadruple.Context.PatternMemberID;
                    this.Commands["INSERT"].Parameters["SUBJ"].Value   = quadruple.Subject.ToString();
                    this.Commands["INSERT"].Parameters["SUBJID"].Value = quadruple.Subject.PatternMemberID;
                    this.Commands["INSERT"].Parameters["PRED"].Value   = quadruple.Predicate.ToString();
                    this.Commands["INSERT"].Parameters["PREDID"].Value = quadruple.Predicate.PatternMemberID;
                    this.Commands["INSERT"].Parameters["OBJ"].Value    = quadruple.Object.ToString();
                    this.Commands["INSERT"].Parameters["OBJID"].Value  = quadruple.Object.PatternMemberID;
                    //Execute command
                    this.Commands["INSERT"].ExecuteNonQuery();
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["INSERT"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot insert data into MySQL store because: " + ex.Message);
                }
            }
            return this;
        }
        #endregion

        #region Remove
        /// <summary>
        /// Removes the given quadruples from the store
        /// </summary>
        public override RDFStore RemoveQuadruple(RDFQuadruple quadruple) {
            if (quadruple != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Open transaction
                    this.Commands["DELETE"].Transaction = this.Commands["DELETE"].Connection.BeginTransaction();
                    //Prepare command
                    this.Commands["DELETE"].Prepare();
                    //Valorize parameters
                    this.Commands["DELETE"].Parameters["QID"].Value = quadruple.QuadrupleID;
                    //Execute command
                    this.Commands["DELETE"].ExecuteNonQuery();
                    //Close transaction
                    this.Commands["DELETE"].Transaction.Commit();
                    //Close connection
                    this.Connection.Close();
                }
                catch (Exception ex) {
                    //Close transaction
                    this.Commands["DELETE"].Transaction.Rollback();
                    //Close connection
                    this.Connection.Close();
                    //Propagate exception
                    throw new RDFStoreException("Cannot delete data from MySQL store because: " + ex.Message);
                }
            }
            return this;
        }

        /// <summary>
        /// Clears the quadruples of the store
        /// </summary>
        public override RDFStore ClearQuadruples() {
            try {
                //Modify command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples";
                this.Commands["DELETE"].Parameters.Clear();
                //Open connection
                this.Connection.Open();
                //Open transaction
                this.Commands["DELETE"].Transaction = this.Commands["DELETE"].Connection.BeginTransaction();
                //Execute command
                this.Commands["DELETE"].ExecuteNonQuery();
                //Close transaction
                this.Commands["DELETE"].Transaction.Commit();
                //Close connection
                this.Connection.Close();
                //Restore command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples WHERE QuadrupleID = @QID";
                this.Commands["DELETE"].Parameters.Add(new MySqlParameter("QID", MySqlDbType.Int64)); 
            }
            catch (Exception ex) {
                //Close transaction
                this.Commands["DELETE"].Transaction.Rollback();
                //Close connection
                this.Connection.Close();
                //Restore command
                this.Commands["DELETE"].CommandText = "DELETE FROM Quadruples WHERE QuadrupleID = @QID";
                this.Commands["DELETE"].Parameters.Add(new MySqlParameter("QID", MySqlDbType.Int64));
                //Propagate exception
                throw new RDFStoreException("Cannot delete data from MySQL store because: " + ex.Message);
            }
            return this;
        }
        #endregion

        #region Select
        /// <summary>
        /// Gets a store containing all quadruples
        /// </summary>
        public override RDFStore SelectAllQuadruples() {
            RDFMemoryStore result                         = new RDFMemoryStore();
            try {
                //Open connection
                this.Connection.Open();
                //Execute command
                using (MySqlDataReader fetchedQuadruples  = this.Commands["SELECT"].ExecuteReader()) {
                    if (fetchedQuadruples.HasRows) {
                        //Iterate fetched quadruples
                        while (fetchedQuadruples.Read()) {
                            //Add the fetched quadruple to the result
                            result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                        }
                    }
                }
                //Close connection
                this.Connection.Close();
            }
            catch (Exception ex) {
                //Close connection
                this.Connection.Close();
                //Propagate exception
                throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified context
        /// </summary>
        public override RDFStore SelectQuadruplesByContext(RDFContext contextResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (contextResource  != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE ContextID = @CTXID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new MySqlParameter("CTXID", MySqlDbType.Int64));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["CTXID"].Value = contextResource.PatternMemberID;
                    //Execute command
                    using (MySqlDataReader fetchedQuadruples          = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified subject
        /// </summary>
        public override RDFStore SelectQuadruplesBySubject(RDFResource subjectResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (subjectResource  != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText               += " WHERE SubjectID = @SUBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new MySqlParameter("SUBJID", MySqlDbType.Int64));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["SUBJID"].Value = subjectResource.PatternMemberID;
                    //Execute command
                    using (MySqlDataReader fetchedQuadruples           = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
								//Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified predicate
        /// </summary>
        public override RDFStore SelectQuadruplesByPredicate(RDFResource predicateResource) {
            RDFMemoryStore result  = new RDFMemoryStore();
            if (predicateResource != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText               += " WHERE PredicateID = @PREDID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new MySqlParameter("PREDID", MySqlDbType.Int64));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["PREDID"].Value = predicateResource.PatternMemberID;
                    //Execute command
                    using (MySqlDataReader fetchedQuadruples           = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText                = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified object
        /// </summary>
        public override RDFStore SelectQuadruplesByObject(RDFResource objectResource) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (objectResource   != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE TripleFlavor = 1 AND ObjectID = @OBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new MySqlParameter("OBJID", MySqlDbType.Int64));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["OBJID"].Value = objectResource.PatternMemberID;
                    //Execute command
                    using (MySqlDataReader fetchedQuadruples          = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Gets a memory store containing quadruples with the specified literal
        /// </summary>
        public override RDFStore SelectQuadruplesByLiteral(RDFLiteral objectLiteral) {
            RDFMemoryStore result = new RDFMemoryStore();
            if (objectLiteral    != null) {
                try {
                    //Open connection
                    this.Connection.Open();
                    //Modify command
                    this.Commands["SELECT"].CommandText              += " WHERE TripleFlavor = 2 AND ObjectID = @OBJID";
                    //Add parameters
                    this.Commands["SELECT"].Parameters.Add(new MySqlParameter("OBJID", MySqlDbType.Int64));
                    //Prepare command
                    this.Commands["SELECT"].Prepare();
                    //Valorize parameters
                    this.Commands["SELECT"].Parameters["OBJID"].Value = objectLiteral.PatternMemberID;
                    //Execute command
                    using (MySqlDataReader fetchedQuadruples          = this.Commands["SELECT"].ExecuteReader()) {
                        if (fetchedQuadruples.HasRows) {
                            //Iterate fetched quadruples
                            while (fetchedQuadruples.Read()) {
                                //Add the fetched quadruple to the result
                                result.AddQuadruple(this.ParseQuadruple(fetchedQuadruples));
                            }
                        }
                    }
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                }
                catch (Exception ex) {
                    //Close connection
                    this.Connection.Close();
                    //Restore command
                    this.Commands["SELECT"].CommandText               = "SELECT TripleFlavor, Context, Subject, Predicate, Object FROM Quadruples";
                    //Restore parameters
                    this.Commands["SELECT"].Parameters.Clear();
                    //Propagate exception
                    throw new RDFStoreException("Cannot read data from MySQL store because: " + ex.Message);
                }
            }
            return result;
        }

        /// <summary>
        /// Parses the current quadruple of the SQLServer reader 
        /// </summary>
        internal RDFQuadruple ParseQuadruple(MySqlDataReader fetchedQuadruples) {
            RDFContext qContext           = new RDFContext(fetchedQuadruples["Context"].ToString());
            RDFResource qSubject          = new RDFResource(fetchedQuadruples["Subject"].ToString());
            RDFResource qPredicate        = new RDFResource(fetchedQuadruples["Predicate"].ToString());
            //SPO-flavour quadruple
            if (fetchedQuadruples["TripleFlavor"].ToString() == "1") {
                RDFResource qObject       = new RDFResource(fetchedQuadruples["Object"].ToString());
                return new RDFQuadruple(qContext, qSubject, qPredicate, qObject);
            }
            //SPL-flavour quadruple
            else {
                String literal            = fetchedQuadruples["Object"].ToString();
                //PlainLiteral
                if (!literal.Contains("^^") || literal.EndsWith("^^") || RDFModelUtilities.GetUriFromString(literal.Substring(literal.LastIndexOf("^^") + 2)) == null) {
                    RDFPlainLiteral pLit  = null;
                    if (literal.Contains("@")) {
                        if (!literal.EndsWith("@")) {
                            Int32 lastAmp = literal.LastIndexOf('@');
                            pLit          = new RDFPlainLiteral(literal.Substring(0, lastAmp), literal.Substring(lastAmp + 1));
                        }
                        else {
                            pLit          = new RDFPlainLiteral(literal);
                        }
                    }
                    else {
                        pLit              = new RDFPlainLiteral(literal);
                    }
                    return new RDFQuadruple(qContext, qSubject, qPredicate, pLit);
                }
                //TypedLiteral
                else {
                    String tLitValue      = literal.Substring(0, literal.LastIndexOf("^^"));
                    String tLitDatatype   = literal.Substring(literal.LastIndexOf("^^") + 2);
                    RDFDatatype dt        = RDFModelUtilities.GetDatatypeFromString(tLitDatatype);
                    RDFTypedLiteral tLit  = new RDFTypedLiteral(tLitValue, dt);
                    return new RDFQuadruple(qContext, qSubject, qPredicate, tLit);
                }
            }
        }
        #endregion

		#region Diagnostics
        /// <summary>
        /// Checks if the MySQL database exists at the given MySQL instance 
        /// </summary>
        internal RDFStoreEnums.RDFStoreSQLErrors MySQLDiagnostics() {
            try {
                //Open connection
                this.Connection.Open();
                //Execute command
                Int32 result = Int32.Parse(this.Commands["DIAGNOSTICS"].ExecuteScalar().ToString());
                //Close connection
                this.Connection.Close();
                //Return the diagnostics state
                if (result  == 0) {
                    return RDFStoreEnums.RDFStoreSQLErrors.QuadruplesTableNotFound;
                }
                else {
                    return RDFStoreEnums.RDFStoreSQLErrors.NoErrors;
                }
            }
            catch {
                //Close connection
                this.Connection.Close();
                //Return the diagnostics state
                return RDFStoreEnums.RDFStoreSQLErrors.InvalidDataSource;
            }
        }
        #endregion

        #region Prepare
        /// <summary>
        /// Creates the default "Quadruples" table
        /// </summary>
        internal void PrepareDatabase() {
            try {
                //Open connection
                this.Connection.Open();
                //Open transaction
                this.Commands["PREPARE"].Transaction = this.Commands["PREPARE"].Connection.BeginTransaction();
                //Execute command
                this.Commands["PREPARE"].ExecuteNonQuery();
                //Close transaction
                this.Commands["PREPARE"].Transaction.Commit();
                //Close connection
                this.Connection.Close();
            }
            catch(Exception ex) {
                //Close transaction
                this.Commands["PREPARE"].Transaction.Rollback();
                //Close connection
                this.Connection.Close();
                //Propagate exception
                throw new RDFStoreException("Cannot prepare MySQL store because: " + ex.Message);
            }
        }
        #endregion		
		
        #endregion

    }

}