﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using RDFSharp.Model;
using RDFSharp.Query;

namespace RDFSharp.Store
{

    /// <summary>
    /// RDFQuadruple represents a quadruple in the RDF store.
    /// </summary>
    public class RDFQuadruple: IEquatable<RDFQuadruple> {

        #region Properties
        /// <summary>
        /// Unique representation of the quadruple
        /// </summary>
        public Int64 QuadrupleID { get; internal set; }

        /// <summary>
        /// Flavor of the triple nested into the quadruple
        /// </summary>
        public RDFModelEnums.RDFTripleFlavors TripleFlavor { get; internal set; }

        /// <summary>
        /// Member acting as context token of the quadruple
        /// </summary>
        public RDFPatternMember Context { get; internal set; }

        /// <summary>
        /// Member acting as subject token of the quadruple
        /// </summary>
        public RDFPatternMember Subject { get; internal set; }

        /// <summary>
        /// Member acting as predicate token of the quadruple
        /// </summary>
        public RDFPatternMember Predicate { get; internal set; }

        /// <summary>
        /// Member acting as object token of the quadruple
        /// </summary>
        public RDFPatternMember Object { get; internal set; }
        #endregion

        #region Ctors
        /// <summary>
        /// SPO-flavor ctor
        /// </summary>
        public RDFQuadruple(RDFContext context, RDFResource subj, RDFResource pred, RDFResource obj) {

            //TripleFlavor
            this.TripleFlavor      = RDFModelEnums.RDFTripleFlavors.SPO;

            //Context
            if (context == null) {
                this.Context       = new RDFContext();
            }
            else {
                this.Context       = context;
            }

            //Subject
            if (subj == null) {
                this.Subject       = new RDFResource();
            }
            else {
                this.Subject       = subj;
            }

            //Predicate
            if (pred != null) {
                if (pred.IsBlank) {
                    throw new RDFStoreException("Cannot create RDFQuadruple because \"pred\" parameter is a blank resource");
                }
                else {
                    this.Predicate = pred;
                }
            }
            else {
                throw new RDFStoreException("Cannot create RDFQuadruple because \"pred\" parameter is null");
            }

            //Object
            if (obj == null) {
                this.Object        = new RDFResource();
            }
            else {
                this.Object        = obj;
            }

            //QuadrupleID
            this.QuadrupleID       = RDFModelUtilities.CreateHash(this.ToString());

        }

        /// <summary>
        /// SPL-flavor ctor
        /// </summary>
        public RDFQuadruple(RDFContext context, RDFResource subj, RDFResource pred, RDFLiteral lit) {

            //TripleFlavor
            this.TripleFlavor      = RDFModelEnums.RDFTripleFlavors.SPL;

            //Context
            if (context == null) {
                this.Context       = new RDFContext();
            }
            else {
                this.Context       = context;
            }

            //Subject
            if (subj == null) {
                this.Subject       = new RDFResource();
            }
            else {
                this.Subject       = subj;
            }

            //Predicate
            if (pred != null) {
                if (pred.IsBlank) {
                    throw new RDFStoreException("Cannot create RDFQuadruple because \"pred\" parameter is a blank resource");
                }
                else {
                    this.Predicate = pred;
                }
            }
            else {
                throw new RDFStoreException("Cannot create RDFQuadruple because \"pred\" parameter is null");
            }

            //Object
            if (lit == null) {
                this.Object        = new RDFPlainLiteral(String.Empty);
            }
            else {
                this.Object        = lit;
            }

            //QuadrupleID
            this.QuadrupleID       = RDFModelUtilities.CreateHash(this.ToString());

        }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the quadruple
        /// </summary>
        public override String ToString() {
            return this.Context + " " + this.Subject + " " + this.Predicate + " " + this.Object;
        }

        /// <summary>
        /// Performs the equality comparison between two quadruples
        /// </summary>
        public Boolean Equals(RDFQuadruple other) {
            if (ReferenceEquals(this, other)) {
                return true;
            }
            if (this      == null) {
                if (other == null) {
                    return true;
                }
                else {
                    return false;
                }
            }
            else {
                if (other == null) {
                    return false;
                }
                else {
                    //Effective Comparison
                    return this.QuadrupleID.Equals(other.QuadrupleID);
                }
            }
        }
        #endregion

    }

}