﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using RDFSharp.Model;

namespace RDFSharp.Store
{

    /// <summary>
    /// RDFStore represents an abstract RDF store, baseline for Memory or SQL-based implementations.
    /// </summary>
    public abstract class RDFStore: IEquatable<RDFStore> {

        #region Properties
        /// <summary>
        /// Unique representation of the store
        /// </summary>
        public Int64 StoreID { get; set; }

        /// <summary>
        /// Type of the store
        /// </summary>
        public String StoreType { get; set; }
        #endregion

        #region Interfaces
        /// <summary>
        /// Gives the string representation of the store
        /// </summary>
        public override String ToString() {
            return this.StoreType;
        }

        /// <summary>
        /// Performs the equality comparison between two stores
        /// </summary>
        public Boolean Equals(RDFStore other) {
            if (ReferenceEquals(this, other)) {
                return true;
            }
            if (this      == null) {
                if (other == null) {
                    return true;
                }
                else {
                    return false;
                }
            }
            else {
                if (other == null) {
                    return false;
                }
                else {
                    //Effective Comparison
                    return this.StoreID.Equals(other.StoreID);
                }
            }
        }
        #endregion

        #region Methods

        #region Add
        /// <summary>
        /// Merges the given graph into the store, avoiding duplicate insertions
        /// </summary>
        public abstract RDFStore MergeGraph(RDFGraph graph);

        /// <summary>
        /// Adds the given quadruple to the store, avoiding duplicate insertions
        /// </summary>
        public abstract RDFStore AddQuadruple(RDFQuadruple quadruple);
        #endregion

        #region Remove
        /// <summary>
        /// Removes the given quadruples from the store
        /// </summary>
        public abstract RDFStore RemoveQuadruple(RDFQuadruple quadruple);

        /// <summary>
        /// Clears the quadruples of the store
        /// </summary>
        public abstract RDFStore ClearQuadruples();
        #endregion

        #region Select
        /// <summary>
        /// Gets a store containing all quadruples
        /// </summary>
        public abstract RDFStore SelectAllQuadruples();

        /// <summary>
        /// Gets a store containing quadruples with the specified context
        /// </summary>
        public abstract RDFStore SelectQuadruplesByContext(RDFContext contextResource);

        /// <summary>
        /// Gets a store containing quadruples with the specified subject 
        /// </summary>
        public abstract RDFStore SelectQuadruplesBySubject(RDFResource subjectResource);

        /// <summary>
        /// Gets a store containing quadruples with the specified predicate
        /// </summary>
        public abstract RDFStore SelectQuadruplesByPredicate(RDFResource predicateResource);

        /// <summary>
        /// Gets a store containing quadruples with the specified object 
        /// </summary>
        public abstract RDFStore SelectQuadruplesByObject(RDFResource objectResource);

        /// <summary>
        /// Gets a store containing quadruples with the specified literal 
        /// </summary>
        public abstract RDFStore SelectQuadruplesByLiteral(RDFLiteral objectLiteral);
        #endregion

        #endregion

    }

}