﻿/*
   Copyright 2012-2013 Marco De Salvo

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;

namespace RDFSharp.Model
{

    /// <summary>
    /// RDFSerializerUtilities is a collector of reusable utility methods for RDF serialization management
    /// </summary>
    internal static class RDFSerializerUtilities {

        #region Methods

        #region N-Triples
        /// <summary>
        /// Regex to catch 8-byte unicodes in N-Triples
        /// </summary>
        private static readonly Regex regexU8 = new Regex(@"\\U([0-9A-Fa-f]{8})");
        /// <summary>
        /// Regex to catch 4-byte unicodes in N-Triples
        /// </summary>
        private static readonly Regex regexU4 = new Regex(@"\\u([0-9A-Fa-f]{4})");
        /// <summary>
        /// Regex to parse N-Triples basing on predicate position 
        /// </summary>
        private static readonly Regex regexNT = new Regex(@"(?<pred> <(.)+> )", RegexOptions.ExplicitCapture); 

        /// <summary>
        /// Tries to parse the given NTriple, throwing an error in case of a basic syntactical error is found 
        /// </summary>
        internal static String[] ParseNTriple(String ntriple) {
            String[] tokens   = new String[3];
            
            //A legal NTriple starts with "_:" of blanks or "<" of non-blanks
            if (ntriple.StartsWith("_:") || ntriple.StartsWith("<")) {
                
                //Parse NTriple by exploiting surrounding spaces and angle brackets of predicate
                tokens        = regexNT.Split(ntriple);
                
                //An illegal NTriple cannot be splitted into three parts with this regex, because predicate not being isolated
                if (tokens.Length != 3) {
                    throw new Exception("found illegal NTriple, predicate must be surrounded by \" <\" and \"> \"");
                }
                else {
                
                    //Check subject for well-formedness
                    tokens[0] = tokens[0].Trim();
                    if (tokens[0].Contains(" ")) {
                        throw new Exception("found illegal NTriple, subject Uri cannot contain spaces");
                    }
                    if ((tokens[0].StartsWith("<")   && !tokens[0].EndsWith(">")) ||
                         (tokens[0].StartsWith("_:") && tokens[0].EndsWith(">"))  ||
                         (tokens[0].Count(c => c.Equals('<')) > 1)                ||
                         (tokens[0].Count(c => c.Equals('>')) > 1))  {
                        throw new Exception("found illegal NTriple, subject Uri is not well-formed");
                    }

                    //Check predicate for well-formedness
                    tokens[1]  = tokens[1].Trim();
                    if (tokens[1].Contains(" ")) {
                        throw new Exception("found illegal NTriple, predicate Uri cannot contain spaces");
                    }
                    if ((tokens[1].Count(c => c.Equals('<')) > 1)                 ||
                        (tokens[1].Count(c => c.Equals('>')) > 1)) {
                        throw new Exception("found illegal NTriple, predicate Uri is not well-formed");
                    }

                    //Check object for well-formedness
                    tokens[2] = tokens[2].Trim();
                    if (tokens[2].StartsWith("<")) {
                        if (tokens[2].Contains(" ")) {
                            throw new Exception("found illegal NTriple, object Uri cannot contain spaces");
                        }
                        if ((!tokens[2].EndsWith(">")                             ||
                             (tokens[2].Count(c => c.Equals('<')) > 1)            ||
                             (tokens[2].Count(c => c.Equals('>')) > 1))) {
                            throw new Exception("found illegal NTriple, object Uri is not well-formed");
                        }
                    }
                    else if (tokens[2].StartsWith("_:")) {
                        if (tokens[2].EndsWith(">")) {
                            throw new Exception("found illegal NTriple, object Uri is not well-formed");
                        }
                    }

                }
            }
            else {
                throw new Exception("found illegal NTriple, must start with \"_:\" or with \"<\"");
            }

            return tokens;
        }

        /// <summary>
        /// <summary>
        /// Turns back ASCII-encoded Unicodes into Unicodes. 
        /// </summary>
        internal static String ASCII_To_Unicode(String asciiString) {
            if (asciiString  != null) {
                asciiString   = regexU8.Replace(asciiString, match => ((Char)Int32.Parse(match.Groups[1].Value, NumberStyles.HexNumber)).ToString());
                asciiString   = regexU4.Replace(asciiString, match => ((Char)Int32.Parse(match.Groups[1].Value, NumberStyles.HexNumber)).ToString());                
            }
            return asciiString;
        }

        /// <summary>
        /// Turns Unicodes into ASCII-encoded Unicodes. 
        /// </summary>
        internal static String Unicode_To_ASCII(String unicodeString) {
            if (unicodeString  != null) {
                StringBuilder b = new StringBuilder();
                Char[] chars    = unicodeString.ToCharArray();
                foreach (Char c in chars) {
                    if (c <= 127) {
                        b.Append(c.ToString());
                    }
                    else {
                        if (c  <= 65535) {
                            b.Append("\\u" + ((Int32)c).ToString("X4"));
                        }
                        else {
                            b.Append("\\U" + ((Int32)c).ToString("X8"));
                        }
                    }
                }
                unicodeString  = b.ToString();
            }
            return unicodeString;
        }
        #endregion

        #region RdfXml
        /// <summary>
        /// Gives the "rdf:RDF" root node of the document
        /// </summary>
        internal static XmlNode GetRdfRootNode(XmlDocument xmlDoc, XmlNamespaceManager nsMgr) {
            XmlNode rdfRDF = xmlDoc.SelectSingleNode(RDFVocabulary.RDF.PREFIX + ":RDF", nsMgr);

            //Try to select "RDF" root node in case of failure
            if (rdfRDF == null) {
                rdfRDF = xmlDoc.SelectSingleNode("RDF", nsMgr);

                //This is not a valid RDF/XML file: the root node is neither "rdf:RDF" or "RDF"!
                //Throw an exception to inform the user about the crappy root node.
                if (rdfRDF == null) {
                    throw new Exception("Given file has not a valid \"rdf:RDF\" or \"RDF\" root node");
                }
            }

            return rdfRDF;
        }

        /// <summary>
        /// Gives the collection of "xmlns" attributes of the "rdf:RDF" root node
        /// </summary>
        internal static XmlAttributeCollection GetXmlnsNamespaces(XmlNode rdfRDF, XmlNamespaceManager nsMgr) {
            XmlAttributeCollection xmlns = rdfRDF.Attributes;
            if (xmlns.Count > 0) {
                IEnumerator iEnum        = xmlns.GetEnumerator();
                while (iEnum.MoveNext()) {
                    XmlAttribute attr    = (XmlAttribute)iEnum.Current;
                    //Search the xmlns in the namespace register and, if not found, add it
                    RDFNamespace ns      =  RDFNamespaceRegister.GetByPrefix(attr.LocalName);
                    if (ns              == null) {
                        ns               = RDFNamespaceRegister.GetByNamespace(attr.Value);
                        if (ns          == null) {
                            ns           = new RDFNamespace(attr.LocalName, attr.Value);
                            //If the namespace has "base" prefix, it must be registered as an automatic namespace
                            if (ns.Prefix.ToLowerInvariant() == "base") {
                                ns       = RDFModelUtilities.GenerateNamespace(ns.ToString(), false);
                            }
                            RDFNamespaceRegister.AddNamespace(ns);
                            nsMgr.AddNamespace(ns.Prefix, ns.Namespace.ToString());
                        }
                    }
                }
            }
            return xmlns;
        }

        /// <summary>
        /// Gives the subj node extracted from the attribute list of the current element 
        /// </summary>
        internal static RDFResource GetSubjectNode(XmlNode subjNode, Uri xmlBase, RDFGraph result) {
            RDFResource subj              = null;

            //If there are attributes, search them for the one representing the subj
            if (subjNode.Attributes      != null && subjNode.Attributes.Count > 0) {                
                //We are interested in finding the "rdf:about" node for the subj
                XmlAttribute rdfAbout     = GetRdfAboutAttribute(subjNode);
                if(rdfAbout != null) {
                    //Attribute found, but we must check if it is "rdf:ID", "rdf:nodeID" or a relative Uri: 
                    //in this case it must be resolved against the xmlBase namespace, or else it remains the same
                    String rdfAboutValue  = RDFSerializerUtilities.ResolveRelativeNode(rdfAbout, xmlBase);
                    subj                  = new RDFResource(rdfAboutValue);
                }

                //If "rdf:about" attribute has been found for the subj, we must
                //check if the node is not a standard "rdf:Description": this is
                //the case we can directly build a triple with "rdf:type" pred
                if (subj != null && !CheckIfRdfDescriptionNode(subjNode)) {
                    RDFResource obj       = new RDFResource(subjNode.NamespaceURI + subjNode.LocalName);
                    RDFTriple     t       = new RDFTriple(subj, RDFVocabulary.RDF.TYPE, obj);
                    result.AddTriple(t);
                }
                
            }

            //There are no attributes, so there's only one way we can handle this element:
            //if it is a standard rdf:Description, it is a blank Subject
            else {
                if (CheckIfRdfDescriptionNode(subjNode)) {
                    subj                  = new RDFResource();
                }                
            }

            return subj;
        }

        /// <summary>
        /// Checks if the given attribute is "rdf:ID", "rdf:nodeID" or a relative Uri: if so, resolves it against the xmlBase namespace
        /// </summary>
        internal static String ResolveRelativeNode(XmlAttribute attr, Uri xmlBase) {
            if (attr != null && xmlBase != null) {
                String attrValue = attr.Value;
                if (attrValue.StartsWith("#")                                       ||    //relative uri starting with "#"
                    RDFModelUtilities.GetUriFromString(attrValue) == null           ||    //relative uri not starting with "#"
                    attr.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":ID")         ||
                    attr.LocalName.Equals("ID")                                     ||
                    attr.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":nodeID")     ||
                    attr.LocalName.Equals("nodeID")) {
                    attrValue = RDFModelUtilities.GetUriFromString(xmlBase.ToString() + "#" + attrValue).ToString();
                }
                return attrValue;
            }
            else {
                throw new RDFModelException("Cannot resolve relative node because \"attr\" or \"xmlBase\" parameter are null");
            }
        }

        /// <summary>
        /// Verify if we are on a standard rdf:Description element
        /// </summary>
        internal static Boolean CheckIfRdfDescriptionNode(XmlNode subjNode) {
            Boolean result = (subjNode.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":Description") ||
                              subjNode.LocalName.Equals("Description"));
            return result;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF subj
        /// </summary>
        internal static XmlAttribute GetRdfAboutAttribute(XmlNode subjNode) {
            //rdf:about
            XmlAttribute rdfAbout         = subjNode.Attributes[RDFVocabulary.RDF.PREFIX + ":about"];
            if (rdfAbout                 == null) {
                //about
                rdfAbout                  = subjNode.Attributes["about"];
                if (rdfAbout             == null) {
                    //rdf:nodeID
                    rdfAbout              = subjNode.Attributes[RDFVocabulary.RDF.PREFIX + ":nodeID"];
                    if (rdfAbout         == null) {
                        //nodeID
                        rdfAbout          = subjNode.Attributes["nodeID"];
                        if (rdfAbout     == null) {
                            //rdf:ID
                            rdfAbout      = subjNode.Attributes[RDFVocabulary.RDF.PREFIX + ":ID"];
                            if (rdfAbout == null) {
                                //ID
                                rdfAbout  = subjNode.Attributes["ID"];
                            }
                        }
                    }
                }
            }
            return rdfAbout;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF object
        /// </summary>
        internal static XmlAttribute GetRdfResourceAttribute(XmlNode predNode) {
            XmlAttribute rdfResource = predNode.Attributes[RDFVocabulary.RDF.PREFIX + ":resource"];
            //rdf:Resource
            if (rdfResource         == null) {
                //resource
                rdfResource          = predNode.Attributes["resource"];
                //rdf:nodeID
                if (rdfResource     == null) {
                    rdfResource      = predNode.Attributes[RDFVocabulary.RDF.PREFIX + ":nodeID"];
                    if (rdfResource == null) {
                        //nodeID
                        rdfResource  = predNode.Attributes["nodeID"];
                    }
                }
            }
            return rdfResource;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF typed literal datatype
        /// </summary>
        internal static XmlAttribute GetRdfDatatypeAttribute(XmlNode predNode) {
            XmlAttribute rdfDatatype = predNode.Attributes[RDFVocabulary.RDF.PREFIX + ":datatype"];
            //rdf:datatype
            if (rdfDatatype         == null) {
                //datatype
                rdfDatatype          = predNode.Attributes["datatype"];
            }
            return rdfDatatype;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF plain literal language
        /// </summary>
        internal static XmlAttribute GetXmlLangAttribute(XmlNode predNode) {
            XmlAttribute xmlLang     = predNode.Attributes[RDFVocabulary.XML.PREFIX + ":lang"];
            //xml:lang
            if (xmlLang             == null) {
                //lang
                xmlLang              = predNode.Attributes["lang"];
            }
            return xmlLang;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF parseType "Collection"
        /// </summary>
        internal static XmlAttribute GetParseTypeCollectionAttribute(XmlNode predNode) {
            //rdf:parseType="Collection"
            XmlAttribute rdfCollection = predNode.Attributes[RDFVocabulary.RDF.PREFIX + ":parseType"];
            if (rdfCollection         != null && rdfCollection.Value.ToLowerInvariant() == "collection") {
                return rdfCollection;
            }
            //parseType="Collection"
            rdfCollection              = predNode.Attributes["parseType"];
            if (rdfCollection         != null && rdfCollection.Value.ToLowerInvariant() == "collection") {
                return rdfCollection;
            }
            return null;
        }

        /// <summary>
        /// Given an element, return the attribute which can correspond to the RDF parseType "Literal"
        /// </summary>
        internal static XmlAttribute GetParseTypeLiteralAttribute(XmlNode predNode) {
            //rdf:parseType="Literal"
            XmlAttribute rdfLiteral    = predNode.Attributes[RDFVocabulary.RDF.PREFIX + ":parseType"];
            if (rdfLiteral            != null && rdfLiteral.Value.ToLowerInvariant() == "literal") {
                return rdfLiteral;
            }
            //parseType="Literal"
            rdfLiteral                 = predNode.Attributes["parseType"];
            if (rdfLiteral            != null && rdfLiteral.Value.ToLowerInvariant() == "literal") {
                return rdfLiteral;
            }
            return null;
        }

        /// <summary>
        /// Given an attribute representing a RDF collection, iterates on its constituent elements
        /// to build its standard reification triples. 
        /// </summary>
        internal static void ParseCollectionElements(XmlNode predNode, RDFResource subj,
                                                     RDFResource pred, RDFGraph result) {

            //Attach the collection as the blank object of the current pred
            RDFResource  obj              = new RDFResource();
            RDFTriple     t               = new RDFTriple(subj, pred, obj);
            result.AddTriple(t);

            //Iterate on the collection items to reify it
            if (predNode.HasChildNodes) {
                XmlNodeList rdfCollItems  = predNode.ChildNodes;
                IEnumerator elems         = rdfCollItems.GetEnumerator();
                while (elems.MoveNext()) {
                    XmlNode elem          = (XmlNode)elems.Current;
                    
                    //Try to get items as "rdf:about" attributes
                    XmlAttribute elemUri  = GetRdfAboutAttribute(elem);
                    if (elemUri          == null) {
                        //Try to get items as "rdf:resource" attributes
                        elemUri           = GetRdfResourceAttribute(elem);
                    }
                    if (elemUri          != null) {

                        // obj -> rdf:type -> rdf:list
                        RDFTriple t1      = new RDFTriple(obj, RDFVocabulary.RDF.TYPE, RDFVocabulary.RDF.LIST);
                        result.AddTriple(t1);

                        // obj -> rdf:first -> res
                        RDFTriple t2      = new RDFTriple(obj, RDFVocabulary.RDF.FIRST, new RDFResource(elemUri.Value));
                        result.AddTriple(t2);

                        //Last element of a collection must give a triple to a "rdf:nil" object
                        RDFResource newObj= null;
						RDFTriple  t3     = null;
						if (elem         != rdfCollItems.Item(rdfCollItems.Count - 1)) {
                            // obj -> rdf:rest -> newObj
                            newObj        = new RDFResource();                            
                        }
                        else {
                            // obj -> rdf:rest -> rdf:nil
                            newObj        = RDFVocabulary.RDF.NIL;
                        }
                        t3                = new RDFTriple(obj, RDFVocabulary.RDF.REST, newObj);
						result.AddTriple(t3);
                        obj               = newObj;
                        continue;
                    }
                }
            }
        }

        /// <summary>
        /// Given the metadata of a graph and a collection resource, it reconstructs the RDF collection and returns it as a list of nodes
        /// This is needed for building the " rdf:parseType=Collection>" RDF/XML abbreviation goody for collections of resources
        /// </summary>
        internal static List<XmlNode> ReconstructCollection(RDFGraphMetadata rdfGraphMetadata, RDFResource tripleObject, XmlDocument rdfDoc) {
            List<XmlNode> result          = new List<XmlNode>();
            Boolean nilFound              = false;
            XmlNode collElementToAppend   = null;
            XmlAttribute collElementAttr  = null;
            XmlText collElementAttrText   = null;
            //Iterate the elements of the collection until the last one (pointing to next="rdf:nil")
            while (!nilFound) {
                var collElement           = rdfGraphMetadata.Collections[tripleObject];
                collElementToAppend       = rdfDoc.CreateNode(XmlNodeType.Element, RDFVocabulary.RDF.PREFIX + ":Description", RDFVocabulary.RDF.BASE_URI);
                collElementAttr           = rdfDoc.CreateAttribute(RDFVocabulary.RDF.PREFIX + ":about", RDFVocabulary.RDF.BASE_URI);
                collElementAttrText       = rdfDoc.CreateTextNode(collElement.ItemValue.ToString());
                collElementAttr.AppendChild(collElementAttrText);
                collElementToAppend.Attributes.Append(collElementAttr);
                result.Add(collElementToAppend);
                //Verify if this is the last element of the collection (pointing to next="rdf:nil")
                if (collElement.ItemNext.ToString().Equals(RDFVocabulary.RDF.NIL.ToString())) {
                    nilFound              = true;
                }
                else {
                    tripleObject          = collElement.ItemNext as RDFResource;
                }
            }
            return result;
        }

        /// <summary>
        /// Verify if we are on a standard rdf:[Bag|Seq|Alt] element
        /// </summary>
        internal static Boolean CheckIfRdfContainerNode(XmlNode containerNode) {
            Boolean result = (containerNode.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":Bag") || containerNode.LocalName.Equals("Bag") ||
                              containerNode.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":Seq") || containerNode.LocalName.Equals("Seq") ||
                              containerNode.LocalName.Equals(RDFVocabulary.RDF.PREFIX + ":Alt") || containerNode.LocalName.Equals("Alt"));
            return result;
        }

        /// <summary>
        /// Given an element, return the child element which can correspond to the RDF container
        /// </summary>
        internal static XmlNode GetContainerNode(XmlNode predNode) {
            //A container is the first child of the given node and it has no attributes.
            //Its localname must be the canonical "rdf:[Bag|Seq|Alt]", so we check for this.
            if (predNode.HasChildNodes) {
                XmlNode containerNode  = predNode.FirstChild;
                Boolean isRdfContainer = CheckIfRdfContainerNode(containerNode);
                if (isRdfContainer) {
                    return containerNode;
                }
                else {
                    return null;
                }
            }
            else {
                return null;
            }
        }

        /// <summary>
        /// Given an element representing a RDF container, iterates on its constituent elements
        /// to build its standard reification triples. 
        /// </summary>
        internal static void ParseContainerElements(RDFModelEnums.RDFContainerTypes contType, XmlNode container, 
                                                    RDFResource subj, RDFResource pred, RDFGraph result) {

            //Attach the container as the blank object of the current pred
            RDFResource  obj              = new RDFResource();
            RDFTriple     t1              = new RDFTriple(subj, pred, obj);
            result.AddTriple(t1);
            
            //obj -> rdf:type -> rdf:[Bag|Seq|Alt]
            RDFTriple t2                  = null;
            if (contType                 == RDFModelEnums.RDFContainerTypes.Bag) {
                t2                        = new RDFTriple(obj, RDFVocabulary.RDF.TYPE, RDFVocabulary.RDF.BAG);
            }
            else if (contType            == RDFModelEnums.RDFContainerTypes.Seq) {
                t2                        = new RDFTriple(obj, RDFVocabulary.RDF.TYPE, RDFVocabulary.RDF.SEQ);
            }
            else {
                t2 = new RDFTriple(obj, RDFVocabulary.RDF.TYPE, RDFVocabulary.RDF.ALT);
            }
            result.AddTriple(t2);

            //Iterate on the container items to reify it
            if (container.HasChildNodes) {
                Int32 enumCounter         = 0;
                XmlNodeList rdfContItems  = container.ChildNodes;
                IEnumerator elems         = rdfContItems.GetEnumerator();
                List<String> elemVals     = new List<String>();
                while (elems.MoveNext()) {
                    XmlNode elem          = (XmlNode)elems.Current;
                    XmlAttribute elemUri  = GetRdfResourceAttribute(elem);
                    
                    //This is a container of resources
                    //obj -> rdf:_N -> OBJ_URI
                    RDFTriple t3          = null;
                    if (elemUri          != null) {
                        if (elem.LocalName.StartsWith("_")) {
                            //"rdf:Alt" containers do not allow duplicate items
                            if (contType == RDFModelEnums.RDFContainerTypes.Alt) {
                                if (!elemVals.Contains(elemUri.Value)) {
                                    elemVals.Add(elemUri.Value);
                                    t3    = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + elem.LocalName), new RDFResource(elemUri.Value)); //respect the "rdf:_N" order
                                }
                            }
                            //Other container types allow duplicate items
                            else {
                                t3        = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + elem.LocalName), new RDFResource(elemUri.Value)); //respect the "rdf:_N" order
                            }
                        }
                        else {
                            //"rdf:Alt" containers do not allow duplicate items
                            if (contType == RDFModelEnums.RDFContainerTypes.Alt) {
                                if (!elemVals.Contains(elemUri.Value)) {
                                    elemVals.Add(elemUri.Value);
                                    t3    = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + "_" + (++enumCounter)), new RDFResource(elemUri.Value)); //assign the "rdf:_N" order
                                }
                            }
                            //Other container types allow duplicate items
                            else {
                                t3        = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + "_" + (++enumCounter)), new RDFResource(elemUri.Value)); //assign the "rdf:_N" order
                            }
                        }
                    }
                    
                    //This is a container of literals (plain with no language)
                    //obj -> rdf:_N -> LIT_VAL
                    else {
                        if (elem.LocalName.StartsWith("_")) {
                            //"rdf:Alt" containers do not allow duplicate items
                            if (contType == RDFModelEnums.RDFContainerTypes.Alt) {
                                if (!elemVals.Contains(elem.InnerText)) {
                                    elemVals.Add(elem.InnerText);
                                    t3    = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + elem.LocalName), new RDFPlainLiteral(elem.InnerText)); //respect the "rdf:_N" order
                                }
                            }
                            //Other container types allow duplicate items
                            else {
                                t3        = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + elem.LocalName), new RDFPlainLiteral(elem.InnerText)); //respect the "rdf:_N" order
                            }
                        }
                        else {
                            //"rdf:Alt" containers do not allow duplicate items
                            if (contType == RDFModelEnums.RDFContainerTypes.Alt) {
                                if (!elemVals.Contains(elem.InnerText)) {
                                    elemVals.Add(elem.InnerText);
                                    t3    = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + "_" + (++enumCounter)), new RDFPlainLiteral(elem.InnerText)); //assign the "rdf:_N" order
                                }
                            }
                            //Other container types allow duplicate items
                            else {
                                t3        = new RDFTriple(obj, new RDFResource(RDFVocabulary.RDF.BASE_URI + "_" + (++enumCounter)), new RDFPlainLiteral(elem.InnerText)); //assign the "rdf:_N" order
                            }
                        }
                    }
                    result.AddTriple(t3);
                    continue;
                }
            }

        }
        #endregion

        #endregion

    }

}